"""Python wrapper for AWS E-Commerce Serive APIs.

Based upon pyamazon (http://www.josephson.org/projects/pyamazon/) with 
efforts to meet the latest AWS specification.

The Amazon's web APIs specication is described here:
http://www.amazon.com/webservices

You need a Amazon-provided license key to use these services.
Follow the link above to get one.  These functions will look in
several places (in this order) for the license key:
    - the "license_key" argument of each function
    - the module-level LICENSE_KEY variable (call setLicense once to set it)
    - an environment variable called AMAZON_LICENSE_KEY
    - foo would return the python object, XMLfoo returns the DOM object

License: Python Software Foundation License
"""


import os, urllib, string, inspect
from xml.dom import minidom

__author__ = "Kun Xi < kunxi@kunxi.org >"
__version__ = "0.2.0"
__license__ = "Python Software Foundation"


"""Package-wide variables:
"""
LICENSE_KEY = None;
HTTP_PROXY = None
LOCALE = "us"
VERSION = "2007-02-22"

__supportedLocales = {
        None : "webservices.amazon.com",  
        "us" : "webservices.amazon.com",   
        "uk" : "webservices.amazon.co.uk",
        "de" : "webservices.amazon.de",
        "jp" : "webservices.amazon.co.jp",
        "fr" : "webservices.amazon.fr",
        "ca" : "webservices.amazon.ca"
    }

__licenseKeys = (
    (lambda key: key),
    (lambda key: LICENSE_KEY), 
    (lambda key: os.environ.get('AWS_LICENSE_KEY', None))
   )


class AWSException(Exception) : 
    '''Base class for all AWS exceptions'''
    pass

class NoLicenseKey(AWSException) : pass
class BadLocale(AWSException) : pass
# Runtime exception
class ExactParameterRequirement(AWSException): pass
class ExceededMaximumParameterValues(AWSException): pass
class InsufficientParameterValues(AWSException): pass
class InternalError(AWSException): pass
class InvalidEnumeratedParameter(AWSException): pass
class InvalidISO8601Time(AWSException): pass
class InvalidOperationForMarketplace(AWSException): pass
class InvalidOperationParameter(AWSException): pass
class InvalidParameterCombination(AWSException): pass
class InvalidParameterValue(AWSException): pass
class InvalidResponseGroup(AWSException): pass
class InvalidServiceParameter(AWSException): pass
class InvalidSubscriptionId(AWSException): pass
class InvalidXSLTAddress(AWSException): pass
class MaximumParameterRequirement(AWSException): pass
class MinimumParameterRequirement(AWSException): pass
class MissingOperationParameter(AWSException): pass
class MissingParameterCombination(AWSException): pass
class MissingParameters(AWSException): pass
class MissingParameterValueCombination(AWSException): pass
class MissingServiceParameter(AWSException): pass
class NoExactMatches(AWSException): pass
class ParameterOutOfRange(AWSException): pass
class ParameterRepeatedInRequest(AWSException): pass
class RestrictedParameterValueCombination(AWSException): pass
class XSLTTransformationError(AWSException): pass

class Bag : 
    '''Wrapper class for DOM nodes'''
    pass



# Utilities functions

def setLocale(locale):
    """set locale
    
    if unsupported locale is set, BadLocale is raised."""
    global LOCALE
    if not __supportedLocales.has_key(locale):
        raise BadLocale, ("Unsupported locale. Locale must be one of: %s" %
            ', '.join([x for x in __supportedLocales.keys() if x]))
    LOCALE = locale


def getLocale():
    """get locale"""
    return LOCALE


def setLicenseKey(license_key=None):
    """set license key

    license key can come from any number of locations;
    see module docs for search order.
    if no license key is specified, BadLocale is raised."""

    global LICENSE_KEY
    for get in __licenseKeys:
        rc = get(license_key)
        if rc: 
            LICENSE_KEY = rc;
            return;
    raise NoLicenseKey, ("Please get the license key from  http://www.amazon.com/webservices")


def getLicenseKey():
    """get license key

    if no license key is specified, BadLocale is raised."""

    if not LICENSE_KEY:
        raise NoLicenseKey, ("Please get the license key from  http://www.amazon.com/webservices")
    return LICENSE_KEY
    

def getVersion():
    """get version"""
    return VERSION


def buildRequest(argv):
    """Build the REST request URL from argv,
    
    all key, value pairs in argv are quoted."""

    url = "http://" + __supportedLocales[getLocale()] + "/onca/xml?Service=AWSECommerceService&"
    url = url + '&'.join(['%s=%s' % (k,urllib.quote(str(v))) for (k,v) in argv.items() if v]) 
    return url


def buildException(els):
    """Build the exception from the returned DOM node

    Only the first exception is raised."""

    error = els[0]
    class_name = error.childNodes[0].firstChild.data[4:]
    msg = error.childNodes[1].firstChild.data 
    if class_name == "ECommerceService.NoExactMatches":
        class_name = "NoExactMatches"

    e = globals()[ class_name ](msg)
    return e


def query(url):
    """Send the query url and return the DOM
    
    Exception is raised if there is errors"""
    u = urllib.FancyURLopener(HTTP_PROXY)
    usock = u.open(url)
    dom = minidom.parse(usock)
    usock.close()

    errors = dom.getElementsByTagName('Error')
    if errors:
        e = buildException(errors)
        raise e
    return dom


def rawObject(XMLSearch, arguments, kwItem, plugins=None):
    '''Return a unique object'''

    dom = XMLSearch(** arguments)
    return unmarshal(dom.getElementsByTagName(kwItem).item(0), plugins) 

    
def rawIterator(XMLSearch, arguments, kwItems, plugins=None):
    '''Return list of objects'''

    dom = XMLSearch(** arguments)
    items = unmarshal(dom.getElementsByTagName(kwItems).item(0), plugins, wrappedIterator())
    return items

class wrappedIterator(list):
    '''Return inherited list object, 
    
    we may add more attributes later'''
    
    pass
    

class pagedIterator:
    '''Return a page-based iterator'''

    def __init__(self, XMLSearch, arguments, kwPage, kwItems, plugins=None):
        """XMLSearch: the callback function that returns the DOM
        arguments: the arguments for XMLSearch
        kwPage, kwItems: Tag name of Page, Items to organize the object
        plugins: please check unmarshal
        """
        self.__search = XMLSearch 
        self.__arguments = arguments 
        self.__keywords ={'Page':kwPage, 'Items':kwItems} 
        self.__plugins = plugins
        self.__page = arguments[kwPage] or 1
        self.__index = 0
        dom = self.__search(** self.__arguments)
        self.__items = unmarshal(dom.getElementsByTagName(kwItems).item(0), plugins, wrappedIterator())
        try:
            self.__len = int(dom.getElementsByTagName("TotalResults").item(0).firstChild.data)
        except AttributeError, e:
            self.__len = len(self.__items)

    def __len__(self):
        return self.__len

    def __iter__(self):
        return self

    def next(self):
        if self.__index < self.__len:
            self.__index = self.__index + 1
            return self.__getitem__(self.__index-1)
        else:
            raise StopIteration

    def __getitem__(self, key):
        try:
            num = int(key)
        except TypeError, e:
            raise e

        if num >= self.__len:
            raise IndexError

        page = num / 10 + 1
        index = num % 10
        if page != self.__page:
            self.__arguments[self.__keywords['Page']] = page
            dom = self.__search(** self.__arguments)
            self.__items = unmarshal(dom.getElementsByTagName(self.__keywords['Items']).item(0), self.__plugins, wrappedIterator())
            self.__page = page

        return self.__items[index]


def unmarshal(element, plugins=None, rc=None):
    """Return the Bag object with attributes populated using DOM element

    element: the root of the DOM element we are interested in
    plugins: callback functions to fine-tune the object structure
    rc: parent object, used in the recursive call

    This core function is inspired by Mark Pilgrim (f8dy@diveintomark.org)
    with some enhancement. Each node.tagName is evalued by plugins' callback
    functions:
        
        if plugins['isBypassed'] is true:
            this elment is ignored
        if plugins['isPivoted'] is true:
            this children of this elment is moved to grandparents
            this object is ignored.
        if plugins['isCollective'] is true:
            this elment is mapped to []
        if plugins['isCollected'] is true:
            this children of elment is appended to grandparent
            this object is ignored.
    """

    if(rc == None):
        rc = Bag()

    if(plugins == None):
        plugins = {}

    childElements = [e for e in element.childNodes if isinstance(e, minidom.Element)]

    if childElements:
        for child in childElements:
            key = child.tagName
            if hasattr(rc, key):
                if type(getattr(rc, key)) <> type([]):
                    setattr(rc, key, [getattr(rc, key)])
                setattr(rc, key, getattr(rc, key) + [unmarshal(child, plugins)])
            elif isinstance(child, minidom.Element):
                if plugins.has_key('isPivoted') and plugins['isPivoted'](child.tagName):
                        unmarshal(child, plugins, rc)
                elif plugins.has_key('isBypassed') and plugins['isBypassed'](child.tagName):
                    continue
                elif plugins.has_key('isCollective') and plugins['isCollective'](child.tagName):
                    setattr(rc, key, unmarshal(child, plugins, wrappedIterator([])))
                elif plugins.has_key('isCollected') and plugins['isCollected'](child.tagName):
                    rc.append(unmarshal(child, plugins))
                else:
                    setattr(rc, key, unmarshal(child, plugins))
    else:
        rc = "".join([e.data for e in element.childNodes if isinstance(e, minidom.Text)])
    return rc


    
# User interfaces

def ItemLookup(ItemId, IdType=None, SearchIndex=None, MerchantId=None, Condition=None, DeliveryMethod=None, ISPUPostalCode=None, OfferPage=None, ReviewPage=None, VariationPage=None, ResponseGroup=None, AWSAccessKeyId=None): 
    '''ItemLookup in ECS'''

    argv = inspect.getargvalues(inspect.currentframe())[-1]
    plugins = {'isPivoted': lambda x: x == 'ItemAttributes', 
        'isCollective': lambda x: x == 'Items', 
        'isCollected': lambda x: x == 'Item'}
    return pagedIterator(XMLItemLookup, argv, 'OfferPage', 'Items', plugins)

    
def XMLItemLookup(ItemId, IdType=None, SearchIndex=None, MerchantId=None, Condition=None, DeliveryMethod=None, ISPUPostalCode=None, OfferPage=None, ReviewPage=None, VariationPage=None, ResponseGroup=None, AWSAccessKeyId=None): 
    '''DOM representation of ItemLookup in ECS'''

    Operation = "ItemLookup"
    AWSAccessKeyId = AWSAccessKeyId or LICENSE_KEY
    argv = inspect.getargvalues(inspect.currentframe())[-1]
    return query(buildRequest(argv))


def ItemSearch(Keywords, SearchIndex="Blended", Availability=None, Title=None, Power=None, BrowseNode=None, Artist=None, Author=None, Actor=None, Director=None, AudienceRating=None, Manufacturer=None, MusicLabel=None, Composer=None, Publisher=None, Brand=None, Conductor=None, Orchestra=None, TextStream=None, ItemPage=None, Sort=None, City=None, Cuisine=None, Neighborhood=None, MinimumPrice=None, MaximumPrice=None, MerchantId=None, Condition=None, DeliveryMethod=None, ResponseGroup=None, AWSAccessKeyId=None):  
    '''ItemSearch in ECS'''

    argv = inspect.getargvalues(inspect.currentframe())[-1]
    plugins = {'isPivoted': lambda x: x == 'ItemAttributes',
        'isCollective': lambda x: x == 'Items', 
        'isCollected': lambda x: x == 'Item'}
    return pagedIterator(XMLItemSearch, argv, "ItemPage", 'Items', plugins)


def XMLItemSearch(Keywords, SearchIndex="Blended", Availability=None, Title=None, Power=None, BrowseNode=None, Artist=None, Author=None, Actor=None, Director=None, AudienceRating=None, Manufacturer=None, MusicLabel=None, Composer=None, Publisher=None, Brand=None, Conductor=None, Orchestra=None, TextStream=None, ItemPage=None, Sort=None, City=None, Cuisine=None, Neighborhood=None, MinimumPrice=None, MaximumPrice=None, MerchantId=None, Condition=None, DeliveryMethod=None, ResponseGroup=None, AWSAccessKeyId=None):  
    '''DOM representation of ItemSearch in ECS'''

    Operation = "ItemSearch"
    AWSAccessKeyId = AWSAccessKeyId or LICENSE_KEY
    argv = inspect.getargvalues(inspect.currentframe())[-1]
    return query(buildRequest(argv))


def SimilarityLookup(ItemId, SimilarityType=None, MerchantId=None, Condition=None, DeliveryMethod=None, ResponseGroup=None, AWSAccessKeyId=None):  
    '''SimilarityLookup in ECS'''

    argv = inspect.getargvalues(inspect.currentframe())[-1]
    plugins = {'isPivoted': lambda x: x == 'ItemAttributes',
        'isCollective': lambda x: x == 'Items',
        'isCollected': lambda x: x == 'Item'}
    return rawIterator(XMLSimilarityLookup, argv, 'Items', plugins)


def XMLSimilarityLookup(ItemId, SimilarityType=None, MerchantId=None, Condition=None, DeliveryMethod=None, ResponseGroup=None, AWSAccessKeyId=None):  
    '''DOM representation of SimilarityLookup in ECS'''

    Operation = "SimilarityLookup"
    AWSAccessKeyId = AWSAccessKeyId or LICENSE_KEY
    argv = inspect.getargvalues(inspect.currentframe())[-1]
    return query(buildRequest(argv))


# List Operations

def ListLookup(ListType, ListId, ProductPage=None, ProductGroup=None, Sort=None, MerchantId=None, Condition=None, DeliveryMethod=None, ResponseGroup=None, AWSAccessKeyId=None):  
    '''ListLookup in ECS'''

    argv = inspect.getargvalues(inspect.currentframe())[-1]
    plugins = {'isPivoted': lambda x: x == 'ItemAttributes',
        'isCollective': lambda x: x == 'Lists', 
        'isCollected': lambda x: x == 'List'}
    return pagedIterator(XMLListLookup, argv, 'ProductPage', 'Lists', plugins)


def XMLListLookup(ListType, ListId, ProductPage=None, ProductGroup=None, Sort=None, MerchantId=None, Condition=None, DeliveryMethod=None, ResponseGroup=None, AWSAccessKeyId=None):  
    '''DOM representation of ListLookup in ECS'''

    Operation = "ListLookup"
    AWSAccessKeyId = AWSAccessKeyId or LICENSE_KEY
    argv = inspect.getargvalues(inspect.currentframe())[-1]
    return query(buildRequest(argv))


def ListSearch(ListType, Name=None, FirstName=None, LastName=None, Email=None, City=None, State=None, ListPage=None, ResponseGroup=None, AWSAccessKeyId=None):
    '''ListSearch in ECS'''

    argv = inspect.getargvalues(inspect.currentframe())[-1]
    plugins = {'isPivoted': lambda x: x == 'ItemAttributes',
        'isCollective': lambda x: x == 'Lists', 
        'isCollected': lambda x: x == 'List'}
    return pagedIterator(XMLListSearch, argv, 'ListPage', 'Lists', plugins)


def XMLListSearch(ListType, Name=None, FirstName=None, LastName=None, Email=None, City=None, State=None, ListPage=None, ResponseGroup=None, AWSAccessKeyId=None):
    '''DOM representation of ListSearch in ECS'''

    Operation = "ListSearch"
    AWSAccessKeyId = AWSAccessKeyId or LICENSE_KEY
    argv = inspect.getargvalues(inspect.currentframe())[-1]
    return query(buildRequest(argv))


#Remote Shopping Cart Operations
def CartCreate(Items, Quantities, ResponseGroup=None, AWSAccessKeyId=None):
    '''CartCreate in ECS'''

    argv = inspect.getargvalues(inspect.currentframe())[-1]
    return __cartOperation(XMLCartCreate, argv)


def XMLCartCreate(Items, Quantities, ResponseGroup=None, AWSAccessKeyId=None):
    '''DOM representation of CartCreate in ECS'''

    Operation = "CartCreate"
    AWSAccessKeyId = AWSAccessKeyId or LICENSE_KEY
    argv = inspect.getargvalues(inspect.currentframe())[-1]
    for x in ('Items', 'Quantities'):
        del argv[x]

    __fromListToItems(argv, Items, 'ASIN', Quantities)
    return query(buildRequest(argv))


def CartAdd(Cart, Items, Quantities, ResponseGroup=None, AWSAccessKeyId=None):
    '''CartAdd in ECS'''

    argv = inspect.getargvalues(inspect.currentframe())[-1]
    return __cartOperation(XMLCartAdd, argv)


def XMLCartAdd(Cart, Items, Quantities, ResponseGroup=None, AWSAccessKeyId=None):
    '''DOM representation of CartAdd in ECS'''

    Operation = "CartAdd"
    AWSAccessKeyId = AWSAccessKeyId or LICENSE_KEY
    CartId = Cart.CartId
    HMAC = Cart.HMAC
    argv = inspect.getargvalues(inspect.currentframe())[-1]
    for x in ('Items', 'Cart', 'Quantities'):
        del argv[x]

    __fromListToItems(argv, Items, 'ASIN', Quantities)
    return query(buildRequest(argv))


def CartGet(Cart, ResponseGroup=None, AWSAccessKeyId=None):
    '''CartGet in ECS'''
    argv = inspect.getargvalues(inspect.currentframe())[-1]
    return __cartOperation(XMLCartGet, argv)


def XMLCartGet(Cart, ResponseGroup=None, AWSAccessKeyId=None):
    '''DOM representation of CartGet in ECS'''

    Operation = "CartGet"
    AWSAccessKeyId = AWSAccessKeyId or LICENSE_KEY
    CartId = Cart.CartId
    HMAC = Cart.HMAC
    argv = inspect.getargvalues(inspect.currentframe())[-1]
    del argv['Cart']
    return query(buildRequest(argv))


def CartModify(Cart, Items, Actions, ResponseGroup=None, AWSAccessKeyId=None):
    '''CartModify in ECS'''

    argv = inspect.getargvalues(inspect.currentframe())[-1]
    return __cartOperation(XMLCartModify, argv)


def XMLCartModify(Cart, Items, Actions, ResponseGroup=None, AWSAccessKeyId=None):
    '''DOM representation of CartModify in ECS'''
    Operation = "CartModify"
    AWSAccessKeyId = AWSAccessKeyId or LICENSE_KEY
    CartId = Cart.CartId
    HMAC = Cart.HMAC
    argv = inspect.getargvalues(inspect.currentframe())[-1]
    for x in ('Cart', 'Items', 'Actions'):
        del argv[x]

    __fromListToItems(argv, Items, 'CartItemId', Actions)
    return query(buildRequest(argv))

    
def CartClear(Cart, ResponseGroup=None, AWSAccessKeyId=None):
    '''CartClear in ECS'''
    argv = inspect.getargvalues(inspect.currentframe())[-1]
    return __cartOperation(XMLCartClear, argv)


def XMLCartClear(Cart, ResponseGroup=None, AWSAccessKeyId=None):
    '''DOM representation of CartClear in ECS'''

    Operation = "CartClear"
    AWSAccessKeyId = AWSAccessKeyId or LICENSE_KEY
    CartId = Cart.CartId
    HMAC = Cart.HMAC
    argv = inspect.getargvalues(inspect.currentframe())[-1]
    del argv['Cart']

    return query(buildRequest(argv))


def __fromListToItems(argv, items, id, actions):
    '''Convert list to AWS REST arguments'''

    for i in range(len(items)):
        argv["Item.%d.%s" % (i+1, id)] = getattr(items[i], id);
        action = actions[i]
        if isinstance(action, int):
            argv["Item.%d.Quantity" % (i+1)] = action
        else:
            argv["Item.%d.Action" % (i+1)] = action


def __cartOperation(XMLSearch, arguments):
    '''Generic cart operation'''

    plugins = {'isBypassed': lambda x: x == 'Request',
        'isCollective': lambda x: x in ('CartItems', 'SavedForLaterItems'),
        'isCollected': lambda x: x in ('CartItem', 'SavedForLaterItem') }
    return rawObject(XMLSearch, arguments, 'Cart', plugins)


# Seller Operation
def SellerLookup(Sellers, ResponseGroup=None, AWSAccessKeyId=None):
    '''SellerLookup in AWS'''

    argv = inspect.getargvalues(inspect.currentframe())[-1]
    plugins = {'isBypassed': lambda x: x == 'Request',
        'isCollective': lambda x: x == 'Sellers',
        'isCollected': lambda x: x == 'Seller'}
    return rawIterator(XMLSellerLookup, argv, 'Sellers', plugins)


def XMLSellerLookup(Sellers, ResponseGroup=None, AWSAccessKeyId=None):
    '''DOM representation of SellerLookup in AWS'''

    Operation = "SellerLookup"
    AWSAccessKeyId = AWSAccessKeyId or LICENSE_KEY
    SellerId = ",".join(Sellers)
    argv = inspect.getargvalues(inspect.currentframe())[-1]
    del argv['Sellers']
    return query(buildRequest(argv))


def SellerListingLookup(SellerId, Id, IdType="Listing", ResponseGroup=None, AWSAccessKeyId=None):
    '''SellerListingLookup in AWS

    Notice: although the repsonse includes TotalPage, TotalResults, 
    there is no ListingPage in the request, so we have to use rawIterator
    instead of pagedIterator. Hope Amazaon would fix this inconsistance'''
    
    argv = inspect.getargvalues(inspect.currentframe())[-1]
    plugins = {'isBypassed': lambda x: x == 'Request',
        'isCollective': lambda x: x == 'SellerListings', 
        'isCollected': lambda x: x == 'SellerListing'}
    return rawIterator(XMLSellerListingLookup, argv, "SellerListings", plugins)


def XMLSellerListingLookup(SellerId, Id, IdType="Listing", ResponseGroup=None, AWSAccessKeyId=None):
    '''DOM representation of SellerListingLookup in AWS'''

    Operation = "SellerListingLookup"
    AWSAccessKeyId = AWSAccessKeyId or LICENSE_KEY
    argv = inspect.getargvalues(inspect.currentframe())[-1]
    return query(buildRequest(argv))


def SellerListingSearch(SellerId, Title=None, Sort=None, ListingPage=None, OfferStatus=None, ResponseGroup=None, AWSAccessKeyId=None):
    '''SellerListingSearch in AWS'''

    argv = inspect.getargvalues(inspect.currentframe())[-1]
    plugins = {'isBypassed': lambda x: x == 'Request',
        'isCollective': lambda x: x == 'SellerListings', 
        'isCollected': lambda x: x == 'SellerListing'}
    return pagedIterator(XMLSellerListingSearch, argv, "ListingPage", "SellerListings", plugins)


def XMLSellerListingSearch(SellerId, Title=None, Sort=None, ListingPage=None, OfferStatus=None, ResponseGroup=None, AWSAccessKeyId=None):
    '''DOM representation of SellerListingSearch in AWS'''

    Operation = "SellerListingSearch"
    AWSAccessKeyId = AWSAccessKeyId or LICENSE_KEY
    argv = inspect.getargvalues(inspect.currentframe())[-1]
    return query(buildRequest(argv))


def CustomerContentSearch(Name=None, Email=None, CustomerPage=1, ResponseGroup=None, AWSAccessKeyId=None):
    '''CustomerContentSearch in AWS'''

    argv = inspect.getargvalues(inspect.currentframe())[-1]
    plugins = {'isBypassed': lambda x: x == 'Request',
        'isCollective': lambda x: x in ('Customers', 'CustomerReviews'),
        'isCollected': lambda x: x in ('Customer', 'Review')}
    return rawIterator(XMLCustomerContentSearch, argv, 'Customers', plugins)


def XMLCustomerContentSearch(Name=None, Email=None, CustomerPage=1, ResponseGroup=None, AWSAccessKeyId=None):
    '''DOM representation of CustomerContentSearch in AWS'''

    Operation = "CustomerContentSearch"
    AWSAccessKeyId = AWSAccessKeyId or LICENSE_KEY
    argv = inspect.getargvalues(inspect.currentframe())[-1]
    for x in ('Name', 'Email'):
        if not argv[x]:
            del argv[x]
    return query(buildRequest(argv))


def CustomerContentLookup(CustomerId, ReviewPage=1, ResponseGroup=None, AWSAccessKeyId=None):
    '''CustomerContentLookup in AWS'''

    argv = inspect.getargvalues(inspect.currentframe())[-1]
    plugins = {'isBypassed': lambda x: x == 'Request',
        'isCollective': lambda x: x == 'Customers',
        'isCollected': lambda x: x == 'Customer'}
    return rawIterator(XMLCustomerContentLookup, argv, 'Customers', plugins)


def XMLCustomerContentLookup(CustomerId, ReviewPage=1, ResponseGroup=None, AWSAccessKeyId=None):
    '''DOM representation of CustomerContentLookup in AWS'''

    Operation = "CustomerContentLookup"
    AWSAccessKeyId = AWSAccessKeyId or LICENSE_KEY
    argv = inspect.getargvalues(inspect.currentframe())[-1]
    return query(buildRequest(argv))


# BrowseNode
def BrowseNodeLookup(BrowseNodeId, ResponseGroup=None, AWSAccessKeyId=None):
    '''BrowseNodeLookup in AWS'''

    argv = inspect.getargvalues(inspect.currentframe())[-1]
    plugins = {'isBypassed': lambda x: x == 'Request',
        'isCollective': lambda x: x == 'Children',
        'isCollected': lambda x: x == 'BrowseNode'}
    return rawIterator(XMLBrowseNodeLookup, argv, 'BrowseNodes', plugins)


def XMLBrowseNodeLookup(BrowseNodeId, ResponseGroup=None, AWSAccessKeyId=None):
    '''DOM representation of BrowseNodeLookup in AWS'''
    
    Operation = "BrowseNodeLookup"
    AWSAccessKeyId = AWSAccessKeyId or LICENSE_KEY
    argv = inspect.getargvalues(inspect.currentframe())[-1]
    return query(buildRequest(argv))


# Help
def Help(HelpType, About, ResponseGroup=None, AWSAccessKeyId=None):
    '''Help in AWS'''

    argv = inspect.getargvalues(inspect.currentframe())[-1]
    plugins = {'isBypassed': lambda x: x == 'Request', 
        'isCollective': lambda x: x in ('RequiredParameters', 
            'AvailableParameters', 'DefaultResponseGroups',
            'AvailableResponseGroups'),
        'isCollected': lambda x: x in ('Parameter', 'ResponseGroup') }
    return rawObject(XMLHelp, argv, 'Information', plugins)


def XMLHelp(HelpType, About, ResponseGroup=None, AWSAccessKeyId=None):
    '''DOM representation of Help in AWS'''

    Operation = "Help"
    AWSAccessKeyId = AWSAccessKeyId or LICENSE_KEY
    argv = inspect.getargvalues(inspect.currentframe())[-1]
    return query(buildRequest(argv))


# Transaction
def TransactionLookup(TransactionId, ResponseGroup=None, AWSAccessKeyId=None):
    '''TransactionLookup in AWS'''

    argv = inspect.getargvalues(inspect.currentframe())[-1]
    plugins = {'isBypassed': lambda x: x == 'Request', 
        'isCollective': lambda x: x in ('Transactions', 'TransactionItems', 'Shipments'),
        'isCollected': lambda x: x in ('Transaction', 'TransactionItem', 'Shipment')}
            
    return rawIterator(XMLTransactionLookup, argv, 'Transactions', plugins)
    

def XMLTransactionLookup(TransactionId, ResponseGroup=None, AWSAccessKeyId=None):
    '''DOM representation of TransactionLookup in AWS'''

    Operation = "TransactionLookup"
    AWSAccessKeyId = AWSAccessKeyId or LICENSE_KEY
    argv = inspect.getargvalues(inspect.currentframe())[-1]
    return query(buildRequest(argv))



if __name__ == "__main__" :
    setLicenseKey("YOUR-LICENSE-HERE");
    sll = SellerListingLookup("A3ENSIQ3ZA4FFN", "1106K206331")
    print dir(sll[0])
    



import os
import mutagen.apev2

TYPE = 'mpc'

# the tags are nearly identical, but we do this anyway for consistency
TAG_TRANSLATION = {
    'artist':       'artist',
    'title':        'title',
    'album':        'album',
    'genre':        'genre',
    'track':        'tracknumber',
    'copyright':    'copyright',
    'isrc':         'isrc',
}

# this code taken from quodlibet
try:
    import ctypes
    _libc = ctypes.cdll.LoadLibrary("libc.so.6")
    _mpcdec = ctypes.cdll.LoadLibrary("libmpcdec.so.3")
except (ImportError, OSError):
    _libc = None
else:
    def _get_errno():
        return ctypes.c_int.in_dll(_libc, "errno").value

    mpc_bool_t = ctypes.c_uint8
    mpc_int16_t = ctypes.c_int16
    mpc_int32_t = ctypes.c_int32
    mpc_int64_t = ctypes.c_int64
    mpc_uint16_t = ctypes.c_uint16
    mpc_uint32_t = ctypes.c_uint32
    mpc_streaminfo_off_t = mpc_int32_t

    class _MPCReader(ctypes.Structure):
        _fields_ = [('read', ctypes.c_void_p),
                    ('seek', ctypes.c_void_p),
                    ('tell', ctypes.c_void_p),
                    ('get_size', ctypes.c_void_p),
                    ('canseek', ctypes.c_void_p),
                    # Actually, all the above are function pointers
                    ('data', ctypes.c_void_p)
                    ]

    class _MPCReaderFile(ctypes.Structure):
        _fields_ = [("reader", _MPCReader),
                    ("file", ctypes.c_void_p), # actually FILE*
                    ("file_size", ctypes.c_long),
                    ("is_seekable", mpc_bool_t)]

    class _MPCStreamInfo(ctypes.Structure):
        _fields_ = [("sample_freq", mpc_uint32_t),
                    ("channels", mpc_uint32_t),
                    ("header_position", mpc_streaminfo_off_t),
                    ("stream_version", mpc_uint32_t),
                    ("bitrate", mpc_uint32_t),
                    ("average_bitrate", ctypes.c_double),
                    ("frames", mpc_uint32_t),
                    ("pcm_samples", mpc_int64_t),
                    ("max_band", mpc_uint32_t),
                    ("istereo", mpc_uint32_t), # 'is' is a Python keyword
                    ('ms', mpc_uint32_t),
                    ("block_size", mpc_uint32_t),
                    ("profile", mpc_uint32_t),
                    ("profile_name", ctypes.c_char_p),
                    ("gain_title", mpc_int16_t),
                    ("gain_album", mpc_int16_t),
                    ("peak_title", mpc_uint16_t),
                    ("peak_album", mpc_uint16_t),

                    ("is_true_gapless", mpc_uint32_t),
                    ("last_frame_samples", mpc_uint32_t),
                    ("encoder_version", mpc_uint32_t),
                    ("encoder", ctypes.c_char * 256),
                    ("tag_offset", mpc_streaminfo_off_t),
                    ("total_file_length", mpc_streaminfo_off_t),
                    ]

    _mpcdec.mpc_reader_setup_file_reader.argtypes = [
        ctypes.POINTER(_MPCReaderFile), ctypes.c_void_p]

    _mpcdec.mpc_streaminfo_read.argtypes = [
        ctypes.POINTER(_MPCStreamInfo), ctypes.POINTER(_MPCReader)]

    _mpcdec.mpc_streaminfo_get_length.restype = ctypes.c_double

def get_tag(tagset, tag):
    try:
        return [unicode(t) for t in tagset[tag]]
    except KeyError:
        return []
    
def write_tag(tr):
    try: info = mutagen.apev2.APEv2(tr.io_loc)
    except mutagen.apev2.APENoHeaderError:
        info = mutagen.apev2.APEv2()

    for mpc_tag, tag in TAG_TRANSLATION.iteritems():
        if tr.tags[tag]:
            info[mpc_tag] = tr.tags[tag]

    tag.save(tr.io_loc)

def can_change(tag):
    return tag in TAG_TRANSLATION.values()

def is_multi():
    return True

def fill_tag_from_path(tr):
    try: info = mutagen.apev2.APEv2(tr.io_loc)
    except mutagen.apev2.APENoHeaderError: return

    for mpc_tag, tag in TAG_TRANSLATION.iteritems():
        tr.tags[tag] = get_tag(info, mpc_tag)

    # determine length and bitrate with the code from quodlibet
    if _libc:
        reader = _MPCReaderFile()
        f = _libc.fopen(tr.io_loc, "r")
        if not f: raise OSError(os.strerror(_get_errno()))
        _mpcdec.mpc_reader_setup_file_reader(
            ctypes.pointer(reader), ctypes.c_void_p(f))
        info = _MPCStreamInfo()

        if _mpcdec.mpc_streaminfo_read(
            ctypes.byref(info), ctypes.byref(reader.reader)):
            raise IOError("not a valid Musepack file")

        tr.length = int(
            _mpcdec.mpc_streaminfo_get_length(ctypes.byref(info)))
        tr.bitrate = int(info.average_bitrate)
        _libc.fclose(reader.file)
                   



# $Id$

import xlmisc, gobject, common, time, urllib, path, os
import xml.dom.minidom
from gettext import gettext as _
SCROBBLER_SESSION = None

@common.threaded
def submit_to_scrobbler(exaile, tr):
    if tr.title == '' or tr.artist == '': return
    if tr.type == 'lastfm': return
   
    if not SCROBBLER_SESSION: return
    len = tr.duration
    lt = time.gmtime(time.time())

    date = "%02d-%02d-%02d %02d:%02d:%02d" % (lt[0], lt[1], lt[2],
        lt[3], lt[4], lt[5])

    xlmisc.log('Attempting to submit track %s to last.fm' % tr)
    try:
        SCROBBLER_SESSION(artist_name=tr.artist,
            song_title=tr.title,
            length=int(tr.duration),
            date_played=date,
            album=tr.album)
    except:
        xlmisc.log_exception()
        gobject.idle_add(exaile.status.set_first, 
            _("Failed to submit track to Last.fm."), 3000)

def load_cache():
    """
        Loads the cache that's been saved to disk
    """
    global SCROBBLER_SESSION
    if SCROBBLER_SESSION == None: return
    file_name = path.get_config('lastfm.xml')
    if not os.path.isfile(file_name):
        xlmisc.log('No last.fm cache found.')
        return
    
    doc = xml.dom.minidom.parse(file_name)
    tracks = doc.getElementsByTagName("track")
    cache = []
    
    for track in tracks:
        artist = getElement(track, "artist")
        title = getElement(track, "title")
        album = getElement(track, "album")
        len = getElement(track, "duration")
        date = getElement(track, "date")
        mbid = getElement(track, "mbid")
        
        # When rebuilding the cache we remove the url codes that were added
        # during a previous failed submission
        entry = {'a[%s]' : artist,
                 't[%s]' : title,
                 'l[%s]' : len,
                 'i[%s]' : date,
                 'b[%s]' : album,
                 'm[%s]' : mbid,
                }
        cache.append(entry)
        doc.childNodes[0].removeChild(track)
    file = open(file_name, "w")
    doc.writexml(file)
    file.close()
    SCROBBLER_SESSION.cache = cache
    SCROBBLER_SESSION.post()
        
def getElement(track, tag):
    """
        Returns the element in a track associated with a given tag,
        or an empty string if the element isn't found or is None
    """
    try:
        ret = track.getElementsByTagName(tag)[0].childNodes[0].nodeValue
    except:
        ret = ''
    return ret
            
def write_cache():
    """
        Writes the local cache to disk when exaile is closed
    """
    global SCROBBLER_SESSION
    if SCROBBLER_SESSION == None: return
    file_name = path.get_config('') + "/lastfm.xml"
    try:
        doc = xml.dom.minidom.parse(file_name)
    except:
        xlmisc.log('last.fm submission cache file not found, creating it...')
        file = open(file_name, "w")
        doc = xml.dom.minidom.Document()
        root_element = doc.createElement("cache")
        doc.appendChild(root_element)
        doc.writexml(file)
        file.close()
        doc = xml.dom.minidom.parse(file_name)
            
    for entry in SCROBBLER_SESSION.cache:
        tr = {}
        for k in entry.keys():
            if k[0] == 'a':
                tr[0] = urllib.unquote(entry[k])
            elif k[0] == 't':
                tr[1] = urllib.unquote(entry[k])
            elif k[0] == 'l':
                tr[2] = urllib.unquote(entry[k])
            elif k[0] == 'i':
                tr[3] = urllib.unquote(entry[k])
            elif k[0] == 'b':
                tr[4] = urllib.unquote(entry[k])
            elif k[0] == 'm':
                tr[5] = urllib.unquote(entry[k])
            else:
                xlmisc.log("Unexpected entry in lastfm cache.  Aborting.")
                return
                
        # Create elements to be added to the xml file    
        track_element = doc.createElement("track")
        title_element = doc.createElement("title")
        artist_element = doc.createElement("artist")
        duration_element = doc.createElement("duration")
        date_element = doc.createElement("date")
        album_element = doc.createElement("album")
        mbid_element = doc.createElement("mbid")
        
        # Add the stored track information to its respective element
        title_element.appendChild(doc.createTextNode(tr[1]))
        artist_element.appendChild(doc.createTextNode(tr[0]))
        duration_element.appendChild(doc.createTextNode(tr[2]))
        date_element.appendChild(doc.createTextNode(tr[3]))
        album_element.appendChild(doc.createTextNode(tr[4]))
        mbid_element.appendChild(doc.createTextNode(tr[5]))
        
        # Add track information elements to the parent track element
        track_element.appendChild(title_element)
        track_element.appendChild(artist_element)
        track_element.appendChild(duration_element)
        track_element.appendChild(date_element)
        track_element.appendChild(album_element)
        track_element.appendChild(mbid_element)

        # Append the new track element to the root cache element
        doc.childNodes[0].appendChild(track_element)
        
    file_object = open(file_name, "w")
    doc.writexml(file_object)

def get_scrobbler_session(exaile, username="", password="", new=False): 
    """
        If there is no audio scrobbler session, one is created and returned,
        else the current one is returned
    """
    global SCROBBLER_SESSION

    if (SCROBBLER_SESSION == None or new) and \
        (username != "" and password != ""):
        xlmisc.log('Logging in to audioscrobbler')
        SCROBBLER_SESSION = AudioScrobblerPost(username=username, 
            password=password, client_name='exa')
        SCROBBLER_SESSION.verbose = True
        try:
            SCROBBLER_SESSION.auth()
            xlmisc.log("Login succesful")
        except:
            xlmisc.log_exception()
            gobject.idle_add(exaile.status.set_first,
                _("Error logging into Last.fm."), 3000)
            return None
    return SCROBBLER_SESSION
"""
~~~~~~~~~~~
PyScrobbler
~~~~~~~~~~~

Python Classes for AudioScrobbler
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
A set of python classes for interacting with the Last.fm AudioScrobbler
services.

AudioScrobblerQuery
    AudioScrobblerQuery is used to consume the web services that provide
    similar artists, charts, tags and so on
    
AudioScrobblerPost
    AudioScrobblerPost is used to update your personal play information for
    those of you (like me) who've been stoopid enough to write your own player.
    
AudioScrobblerQuery: AudioScrobbler Web Service Consumption
=============================================================

A set of classes for consuming the AudioScrobbler web services.
Currently we consume protocol 1.0, as defined at:
http://www.audioscrobbler.net/data/webservices

Basic usage
-----------
>>> lips = AudioScrobblerQuery(artist='The Flaming Lips')
>>> for artist in lips.similar():
...     print artist.name, artist.mbid
... 

or

>>> for i in AudioScrobblerQuery(user='offmessage').topalbums()
...     print i.name, i.image.small
... 

A certain number of the services take a querystring (for example a user's
weekly artist chart takes a start and end date).  These can be passed as
follows:

>>> offmessage = AudioScrobblerQuery(user='offmessage')
>>> for i in offmessage.weeklyartistchart(_from=1114965332, to=1115570132):
...     print i.name, i.chartposition
... 

Note that we have had to prefix the ``from`` keyword with an underscore
(``_``). This is because ``from`` is a reserved word in Python and can't be 
used as a parameter name.  Any parameter passed with a leading underscore will 
have that underscore stripped (so ``_from`` is passed as ``from``, as ``_foo`` 
would be passed as ``foo``).

If an element has an attribute (for example the artist has an ``mbid`` in
/user/.../topalbums.xml) these can be accessed as a dictionary key of that
element:

>>> offmessage = AudioScrobblerQuery(user='offmessage')
>>> for i in offmessage.topalbums():
...     print i.name, i.artist, i.artist['mbid']
... 

if you're not sure if an element has an attribute they can be accessed instead
using the ``get()`` method:

>>> offmessage = AudioScrobblerQuery(user='offmessage')
>>> for i in offmessage.topalbums():
...     print i.name, i.artist, i.artist.get('mbid', '')
... 

More control
------------
There are 2 additional keyword arguments that you can offer at the point of
instantiation of the AudioScrobblerQuery object:

host
    If the lovely people at Last.fm change their server name, or if you set up
    some kind of local proxy of their services you can set the host from which
    you get the XML.
    
    Usage:
    >>> lips = AudioScrobblerQuery(artist='The Flaming Lips', 
    ...                              host='audioscrobbler.isotoma.com')
    
    Default is ``'ws.audioscrobbler.com'``
    
version
    If another version of the web services is released, and this code still
    holds up you can set the protocol version at instantiation.  This is 
    inserted into the request URL at the correct point (i.e. 
    ws.audioscrobbler.com/**1.0**/artist/The+Flaming+Lips).
    
    Usage:
    >>> lips = AudioScrobberRequest(artist='The Flaming Lips', version='1.1')
    
    Default is ``'1.0'``

Caching
-------
Each instance of the ``AudioScrobblerQuery`` maintains its own internal cache
of requests, which are checked against the server's response.  If the server
returns a ``304`` (content not modified) we return the cached value, otherwise
we return the result that the server gave us.

This is all well and good, however at the time of writing the
``ws.audioscrobbler.com`` server does not support the ``304`` response for all
pages, so you may find yourself making requests that you are sure should be
cached, but in fact aren't.

For example:
http://ws.audioscrobbler.com/1.0/user/offmessage/topalbums.xml returns a ``304``
http://ws.audioscrobbler.com/1.0/artist/The+Flaming+Lips/similar.xml does not

I have assumed that this is for a reason, and as such honour this behaviour.
If a page does not return a ``Last-Modified`` header we do not cache it.  Any
page that does return such a header (but not the ``Pragma: no-cache`` header)
gets cached.  Any page that does return a ``Last-Modified`` header gets exactly
what it sent to us sent back to it as the ``If-Modified-Since``.  This is a fit
of pragmatism based on the fact that calculating local dates and avoiding
clock disparaties is not worth the effort.

Also, note that the AudioScrobbler servers do not return the ``ETag`` header 
and so we do not look for it or refer to it.

A last note
-----------
These classes are for consuming the AudioScrobbler web services.  As such each
of the classes here have no ``set``, ``__setitem__`` or ``__setattr__``
methods.  They are read only.  If you want to actually manipulate the
resulting XML you should access the underlying ElementTree_ instances directly
using either the ``raw()`` or ``element()`` methods provided.  You can then 
make use of the innate power of those classes directly:

``raw()`` returns an ElementTree:

>>> tree = AudioScrobblerQuery(user='offmessage').topalbums().raw()

while ``element()`` returns an Element at the point within the tree that you
currently are:

>>> lips = AudioScrobblerQuery(artist='The Flaming Lips')
>>> elements = [ x.name.element() for x in lips.similar() ]

which is equivalent to:

>>> lips = AudioScrobberRequest(artist='The Flaming Lips')
>>> elements = lips.similar().raw().findall('*/name')

As an alternative you could also look at XMLTramp_, which is another XML
manipulator designed primarily for consuming web services (and used in the
FlickrClient_ script that gave me much inspiration for these classes).

Last but not least a mention should go to BeautifulSoup_, just because it is
the most powerful and flexible of this type of application that I've used.  It
specialises in badly formed XML/HTML though, and luckily is not required in
this instance.  That said, it's a powerful beast and you should have a look at
it the minute your requirements become more complex than this code supports.

.. _ElementTree: http://effbot.org/zone/element.htm
.. _XMLTramp: http://www.aaronsw.com/2002/xmltramp/
.. _BeautifulSoup: http://www.crummy.com/software/BeautifulSoup/


AudioScrobblerPost: Update Your Personal Play History
=====================================================

A set of classes for posting playlist information to your Last.fm account.
Please note that you will need your own account to use these classes.

Some stuff about the time management and internal cache should go here.

Basic usage
-----------
>>> track = dict(artist_name="The Flaming Lips",
                 song_title="The Yeah Yeah Yeah Song",
                 length=291,
                 date_played="2006-04-29 04:20:00", 
                 album="At War With The Mystics",
                 mbid=""
                )
>>> post = AudioScrobblerPost(user='offmessage', password='mypasswd')
>>> post(**track)
>>> 

More control
------------
There are a number of keywords that can be passed at the time of instantiation
of the ``AudioScrobblerPost`` class:

verbose
    Boolean.  If you pass ``verbose=True`` every log line is also printed.
    Useful for debugging and in interactive scripts.
    
client_name
    By default pyscrobbler uses the ``tst`` client name.  If you deploy these
    classes in an application with a wide user base it seems to be advisable to
    apply for your own client name.  If lots of people start to use these
    classes as is I will apply for a unique name for them.  If you have your
    own, pass it in here (as a string)
    
client_version
    Currently we use ``1.0``.  This is because we are using the ``tst`` client
    name (see above) and anything lower seems to raise an ``UPDATE`` message
    from the post server at handshake.
    
protocol_version
    By default we use ``1.1``.  These classes are not compatible with Protocol
    1.0.  It's unlikely they will be compatible with any later versions.  The
    changes between 1.0 and 1.1 were pretty major; I can't see they'd be any
    less major for the next version.
    
host
    By default we use ``post.audioscrobbler.com``.  If the server name changes
    (or you want to run tests) you can pass a different host name here.  In
    fact, the url that we authenticate against is built as
    ``"http://%s/?%s" % (host, params)`` so as long as your host string makes
    a valid url in that format, go for it.  Note the trailing slash between
    host and the query string.  This seems to be a requirement of urllib2.  This
    means that your test auth script should be able to be called that way.

Putting them to use in real life
================================
Some advice on using these classes.

``timeout``
-----------
I've not done anything about the socket timeout on the running system.  It's
not really this code's place to do anything about it, but it is something that
probably needs sorting out, otherwise your scripts will hang for as long as your 
timeout setting if there is a problem with the Last.fm servers (which there
appears to be fairly frequently with the submissions servers).

Generally, if you are on a slow link or one with poor latency you can up the 
timeout so that your requests don't keep failing unnecessarily.  However, if
you're on a quick link you may want to lower it as low as 10 seconds, as the
wait can be agonising if you're making a few requests and the server is not
responding.

Thanks
======
I took a considerable amount of inspiration from Michael Campeotto's
FlickrClient_ script for AudioScrobblerQuery

I've taken even greater amounts of inspiration from Mike Bleyer's
iPodScrobbler_ script for AudioScrobblerPost.

.. _FlickrClient: http://micampe.it/things/flickrclient
.. _iPodScrobbler: http://www.hoc.net/mike/source/iPodScrobbler/


TODO
====
Mad leet reStructuredText epydoc mashup
    Get that sorted.

Example query script
    Actually write some example query scripts :)
    
Document use of the factory class
    AudioScrobbler
    
"""

__author__ = "Andy Theyers <andy@isotoma.com>"
__version__ = "$Revision$"[11:-2]
__docformat__ = "restructuredtext"


import datetime, locale, md5, pickle, re, site, sys, time, urllib, urllib2

try:
    from xml.etree.ElementTree import ElementTree
except ImportError:
    from elementtree.ElementTree import ElementTree

# This is lifted in the most part from iPodScrobbler (see docs above)
# Get the base local encoding
enc = 'ascii'
try:
    enc = site.encoding
except:
    pass
if enc == 'ascii' and locale.getpreferredencoding():
    enc = locale.getpreferredencoding()
# if we are on MacOSX, we default to UTF8
# because apples python reports 'ISO8859-1' as locale, but MacOSX uses utf8
if sys.platform=='darwin':
    enc = 'utf8'

# AudioScrobblerQuery configuration settings
audioscrobbler_request_version = '1.0'
audioscrobbler_request_host = 'ws.audioscrobbler.com'

# AudioScrobblerPost configuration settings
audioscrobbler_post_version = u'1.1'
audioscrobbler_post_host = 'post.audioscrobbler.com'
client_name = u'exa'
pyscrobbler_version = u'1.0' # This is set to 1.0 while we use
                             # client_name = u'tst' as we keep getting
                             # UPDATE responses with anything less.

class AudioScrobblerError(Exception):
    
    """
    The base AudioScrobbler error from which all our other exceptions derive
    """
    
    def __init__(self, message):
        """ Create a new AudioScrobblerError.

            :Parameters:
                - `message`: The message that will be displayed to the user
        """
        self.message = message
    
    def __repr__(self):
        msg = "%s: %s"
        return msg % (self.__class__.__name__, self.message,)
    
    def __str__(self):
        return self.__repr__()

class AudioScrobblerConnectionError(AudioScrobblerError):
    
    """
    The base network error, raise by invalid HTTP responses or failures to
    connect to the server specified (DNS, timeouts, etc.)
    """
    
    def __init__(self, type, code, message):
        self.type = type
        self.code = code
        self.message = message
    
    def __repr__(self):
        msg = "AudioScrobblerConnectionError: %s: %s %s"
        return msg % (self.type.upper(), self.code, self.message,)
    
class AudioScrobblerTypeError(AudioScrobblerError):
    
    """
    If we would normally raise a TypeError we raise this instead
    """
    pass
    

class AudioScrobblerPostUpdate(AudioScrobblerError):
    
    """
    If the POST server returns an ``UPDATE`` message we raise this exception.
    This is the only case when this exception is raised, allowing you to
    happily ignore it.
    """
    
    def __repr__(self):
        msg = "An update to your AudioScrobbler client is available at %s"
        return msg % (self.message,)


class AudioScrobblerPostFailed(AudioScrobblerError):
    
    """
    If the POST server returns an ``FAILED`` message we raise this exception.
    """
    
    def __repr__(self):
        msg = "Posting track to AudioScrobbler failed.  Reason given: %s"
        return msg % (self.message,)


class AudioScrobblerHandshakeError(AudioScrobblerError):
    
    """
    If we fail the handshake this is raised.  If you're running in a long
    running process you should pass on this error, as the system keeps a
    cache which ``flushcache()`` will clear for you when the server is back.
    """
    pass

class AudioScrobbler:
    
    """ Factory for Queries and Posts.  Holds configuration for the session """
    
    def __init__(self,
                 audioscrobbler_request_version=audioscrobbler_request_version,
                 audioscrobbler_post_version=audioscrobbler_post_version,
                 audioscrobbler_request_host=audioscrobbler_request_host,
                 audioscrobbler_post_host=audioscrobbler_post_host,
                 client_name=client_name,
                 client_version=pyscrobbler_version):
                 
        self.audioscrobbler_request_version=audioscrobbler_request_version
        self.audioscrobbler_post_version=audioscrobbler_post_version
        self.audioscrobbler_request_host=audioscrobbler_request_host
        self.audioscrobbler_post_host=audioscrobbler_post_host
        self.client_name=client_name
        self.client_version=pyscrobbler_version
        
    def query(self, **kwargs):
        
        """ Create a new AudioScrobblerQuery """
        
        if len(kwargs) != 1:
            raise TypeError("__init__() takes exactly 1 argument, %s "
                            "given" % (len(kwargs),))
        ret = AudioScrobblerQuery(version=self.audioscrobbler_request_version,
                                  host=self.audioscrobbler_request_host,
                                  **kwargs)
        return ret
        
    def post(self, username, password, verbose=False):
        
        """ Create a new AudioScrobblerPost """
        
        ret = AudioScrobblerPost(username=username.encode('utf8'),
                                 password=password.encode('utf8'),
                                 host=self.audioscrobbler_post_host,
                                 protocol_version=self.audioscrobbler_post_version,
                                 client_name=self.client_name,
                                 client_version=self.client_version,
                                 verbose=verbose)
        return ret
        

class AudioScrobblerCache:
    def __init__(self, elemtree, last):
        self.elemtree = elemtree
        self.requestdate = last
        
    def created(self):
        return self.requestdate
        
    def gettree(self):
        return self.elemtree

        
class AudioScrobblerQuery:
    
    def __init__(self,
                 version=audioscrobbler_request_version, 
                 host=audioscrobbler_request_host, 
                 **kwargs):
        
        if len(kwargs) != 1:
            raise TypeError("__init__() takes exactly 1 audioscrobbler "
                            "request argument, %s given" % str(len(kwargs))
                           )
        self.type = kwargs.keys()[0]
        self.param = str(kwargs[self.type])
        self.baseurl = 'http://%s/%s/%s/%s' % (host, 
                                               version, 
                                               urllib.quote(self.type), 
                                               urllib.quote(self.param), 
                                              )
        self._cache = {}
        
    def __getattr__(self, name):
        def method(_self=self, name=name, **params):
            # Build the URL
            url = '%s/%s.xml' % (_self.baseurl, urllib.quote(name))
            if len(params) != 0:
                for key in params.keys():
                    # This little mess is required to get round the fact that
                    # 'from' is a reserved word and can't be passed as a
                    # parameter name.
                    if key.startswith('_'):
                        params[key[1:]] = params[key]
                        del params[key]
                url = '%s?%s' % (url, urllib.urlencode(params))
            req = urllib2.Request(url)
            
            # Check the cache
            cache = _self._cache
            if url in cache and cache[url].created() is not None:
                req.add_header('If-Modified-Since', cache[url].created())
            
            # Open the URL and test the response
            try:
                response = urllib2.urlopen(url)
            except urllib2.HTTPError, error:
                if error.code == 304:
                    return AudioScrobblerItem(cache[url].getroot(), _self, url)
                if error.code == 400:
                    raise AudioScrobblerConnectionError('ws', 400, error.fp.read())
                raise AudioScrobblerConnectionError('http', error.code, error.msg)
            except urllib2.URLError, error:
                code = error.reason.args[0]
                message = error.reason.args[1]
                raise AudioScrobblerConnectionError('network', code, message)
            elemtree = ElementTree(file=response)
            if response.headers.get('pragma', None) != 'no-cache':
                last_modified = response.headers.get('last-modified', None)
            else:
                last_modified = None
            _self._cache[url] = AudioScrobblerCache(elemtree, last_modified)
            return AudioScrobblerItem(elemtree.getroot(), _self, url)
        return method
        
    def __repr__(self):
        return "AudioScrobblerQuery(%s='%s')" % (self.type, self.param)
        
    def __str__(self):
        return self.__repr__()

        
class AudioScrobblerItem:
    
    def __init__(self, element, parent, url=None):
        self._element = element
        self._parent = parent
        self.tag = element.tag
        self.text = element.text
        self._url = url
        if self._url is None:
            self._url = self._parent._url
    
    def __repr__(self):
        if isinstance(self._parent, AudioScrobblerQuery):
            return "<AudioScrobbler response from %s>" % (self._url,)
        return "<AudioScrobbler %s element at %s>" % (self.tag, id(self))
        
    def __str__(self):
        if self.text is None:
            return ''
        text = self.text
        try:
            retval = text.encode(enc)
        except AttributeError:
            retval = text
        return retval
    
    def __getattr__(self, name):
        result = self._element.findall(name)
        if len(result) == 0:
            raise AttributeError("AudioScrobbler %s element has no "
                                 "subelement '%s'" % (self.tag, name)
                                )
        ret = [ AudioScrobblerItem(i, self) for i in result ]
        if len(ret) == 1:
            return ret[0]
        return ret
    
    def __iter__(self):
        for i in self._element:
            yield AudioScrobblerItem(i, self)
            
    def __getitem__(self, i):
        return self._element.attrib[i]
        
    def get(self, i, default):
        if i in self._element.attrib:
            return self._element.attrib[i]
        else:
            return default
        
    def __getslice__(self, i, j):
        return [ AudioScrobblerItem(x, self) for x in self._element[i:j] ]
    
    def raw(self):
        def getparent(obj):
            if isinstance(obj._parent, AudioScrobblerQuery):
                return obj._parent
            return getparent(obj._parent)
        return getparent(self)._cache[self._url].gettree()
    
    def element(self):
        return self._element


class AudioScrobblerPost:
    
    """
    Provide the ability to post tracks played to a user's Last.fm
    account
    """
    
    def __init__(self,
                 username=u'',
                 password=u'',
                 client_name=client_name,
                 client_version=pyscrobbler_version,
                 protocol_version=audioscrobbler_post_version,
                 host=audioscrobbler_post_host,
                 verbose=False):
    
        # Capture the information passed for future use
        self.params = dict(username=username,
                           password=password,
                           client_name=client_name,
                           client_version=client_version,
                           protocol_version=protocol_version,
                           host=host)
        
        self.verbose = verbose
        
        self.auth_details = {}
        self.last_post = None
        self.interval = 0
        self.cache = []
        self.loglines = []
        self.last_shake = None
        self.authenticated = False
        self.updateurl = None
        self.posturl = None
        
    def auth(self):
        
        """ Authenticate against the server """
        
        if self.authenticated:
            return True
        now = datetime.datetime.utcnow()
        interval = datetime.timedelta(minutes=30)
        if self.last_shake is not None and self.last_shake + interval > now:
            last_shake_string = self.last_shake.strftime("%Y-%m-%d %H:%M:%S")
            msg = ("Tried to reauthenticate too soon after last try "
                   "(last try at %s)" % (last_shake_string,))
            raise AudioScrobblerHandshakeError(msg)
        
        p = {}
        p['hs'] = 'true'
        p['u'] = self.params['username']
        p['c'] = self.params['client_name']
        p['v'] = self.params['client_version']
        p['p'] = self.params['protocol_version']
        plist = [(k, urllib.quote_plus(v.encode('utf8'))) for k, v in p.items()]
        
        authparams = urllib.urlencode(plist)
        url = 'http://%s/?%s' % (self.params['host'], authparams)
        req = urllib2.Request(url)
        try:
            url_handle = urllib2.urlopen(req)
        except urllib2.HTTPError, error:
            self.authenticated = False
            raise AudioScrobblerConnectionError('http', error.code, error.msg)
        except urllib2.URLError, error:
            self.authenticated = False
            code = error.reason.args[0]
            message = error.reason.args[1]
            raise AudioScrobblerConnectionError('network', code, message)
        
        self.last_shake = datetime.datetime.utcnow()
        response = url_handle.readlines()
        if len(response) == 0:
            raise AudioScrobblerHandshakeError('Got nothing back from the server')
        
        # Get the interval if there is one
        find_interval = re.match('INTERVAL (\d+)', response[-1])
        if find_interval is not None:
            self.interval = int(find_interval.group(1))
            
        username = self.params['username']
        password = self.params['password']
        # First we test the best and most likely case
        if response[0].startswith('UPTODATE'):
            ask = response[1].strip()
            answer = md5.md5(md5.md5(password).hexdigest() + ask).hexdigest()
            self.auth_details['u'] = urllib.quote_plus(username.encode('utf8'))
            self.auth_details['s'] = answer
            self.posturl = response[2].strip()
            self.authenticated = True
            
        # Next we test the least significant failure.
        elif response[0].startswith('UPDATE'):
            updateurl = result[0][7:].strip()
            
            ask = response[1].strip()
            answer = md5.md5(md5.md5(password).hexdigest() + ask).hexdigest()
            
            self.auth_details['u'] = urllib.quote_plus(username.encode('utf8'))
            self.auth_details['s'] = answer
            self.posturl = response[2].strip()
            self.authenticated = True
            self.updateavailable = updateurl
            
        # Then the various failure states....
        elif response[0].startswith('BADUSER'):
            self.authenticated = False
            msg = "Username '%s' unknown by Last.fm"
            raise AudioScrobblerHandshakeError(msg % (username,))
            
        elif response[0].startswith('FAILED'):
            self.authenticated = False
            reason = response[0][6:]
            msg = reason.decode('utf8').encode(enc)
            raise AudioScrobblerHandshakeError(msg)
            
        else:
            self.authenticated = False
            msg = "Unknown response from server: %r" % (response,)
            raise AudioScrobblerHandshakeError(msg)
            
    def posttrack(self, 
                  artist_name,
                  song_title,
                  length,
                  date_played, 
                  album=u'',
                  mbid=u''):
        
        """
        Add the track to the local cache, and try to post it to the server
        """
        
        self.addtrack(artist_name=artist_name,
                      song_title=song_title,
                      length=length,
                      date_played=date_played,
                      album=album,
                      mbid=mbid)
        self.post()
        
    __call__ = posttrack
    
    def addtrack(self, 
                 artist_name,
                 song_title,
                 length,
                 date_played, 
                 album=u'',
                 mbid=u''):
        
        """ Add a track to the local cache """
        
        # Quick sanity check on track length
        if type(length) == type(0):
            sane_length = length
        elif type(length) == type('') and length.isdigit():
            sane_length = int(length)
        else:
            sane_length = -1
        # If the track length is less than 30 move on
        if sane_length < 30:
            msg = ("Track '%s' by '%s' only %s seconds in length, so "
                   "not added" % (song_title, artist_name, sane_length))
            self.log(msg)
        # Otherwise build the track dictionary and add it to the local cache
        else:
            track = {'a[%s]': artist_name.decode(enc).encode('utf8'),
                     't[%s]': song_title.decode(enc).encode('utf8'),
                     'l[%s]': str(sane_length),
                     'i[%s]': date_played,
                     'b[%s]': album.decode(enc).encode('utf8'),
                     'm[%s]': mbid.encode('utf8'),
                    }
            self.cache.append(track)
        
    def post(self):
        
        """
        Post the tracks by popping the first ten off the cache and attempting
        to post them.
        """
        
        if len(self.cache) == 0:
            return
        if len(self.cache) > 10:
            number = 10
        else:
            number = len(self.cache)
        
        params = {}
        count = 0
        for track in self.cache[:number]:
            for k in track.keys():
                track[k] = urllib.unquote(track[k])
                track[k] = urllib.quote(track[k].encode('utf-8'))
                params[k % (count,)] = track[k]
            count += 1
        
        self.auth()
        params.update(self.auth_details)

        postdata = '&'.join(['%s=%s' % (k, v) for k, v in params.iteritems()])
#        postdata = urllib.urlencode(params)
        req = urllib2.Request(url=self.posturl, data=postdata)
        
        now = datetime.datetime.utcnow()
        interval = datetime.timedelta(seconds=self.interval)
        if self.last_post is not None and self.last_post + interval > now:
            time.sleep(self.interval)

        xlmisc.log('-AUDIOSCROBBLER- %s' % postdata)
        try:
            url_handle = urllib2.urlopen(req)
        except urllib2.HTTPError, error:
            raise AudioScrobblerConnectionError('http', error.code, error.msg)
        except urllib2.URLError, error:
            args = getattr(error.reason, 'args', None)
            code = '000'
            message = str(error)
            if args is not None:
                if len(args) == 1:
                    message = error.reason.args[0]
                elif len(args) == 2:
                    code = error.reason.args[0]
                    message = error.reason.args[1]
            raise AudioScrobblerConnectionError('network', code, message)
        
        self.last_post = now
        response = url_handle.readlines()
        
        # Get the interval if there is one
        find_interval = re.match('INTERVAL (\d+)', response[-1])
        if find_interval is not None:
            self.interval = int(find_interval.group(1))
            
        # Test the various responses possibilities:
        if response[0].startswith('OK'):
            self.log("Uploaded %s tracks successfully" % (number,))
            del self.cache[:number]
        elif response[0].startswith('BADAUTH'):
            self.log("Got BADAUTH")
            self.authenticated = False
        elif response[0].startswith('FAILED'):
            reason = response[0][6:].strip()
            raise AudioScrobblerPostFailed(reason)
        else:
            msg = "Server returned something unexpected"
            raise AudioScrobblerPostFailed(msg)
        
    def flushcache(self):
        
        """ Post all the tracks in the cache """
        
        while len(self.cache) > 0:
            self.post()
        
    def log(self, msg):
        
        """ Add a line to the log, print it if verbose """
        
        time = datetime.datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S")
        self.loglines.append("%s: %s" % (time, msg))
        if self.verbose:
            print self.loglines[-1]
            
    def getlog(self, clear=False):
        
        """ Return the entire log, clear it if requested """
        
        if clear:
            retval = self.loglines
            self.loglines = []
            return retval
        return self.loglines



#!/usr/bin/env python

import urllib, sys, re

reg = re.compile(r'<tr class="\w+ prio\d+"><td class="id">.*?class="summary">'
    '<a href="/trac/ticket/(\d+)" title="View ticket">(.*?)</a></td>.*?<td '
    'class="type"><span>(\w+)</span></td>', re.DOTALL)

version = sys.argv[1]
type = sys.argv[2]

data = urllib.urlopen('http://www.exaile.org/trac/query'
    '?status=closed&milestone=%s' % version).read()
data = data.replace('&#34;', '"')

items = reg.findall(data)
items.sort()

for item in items:
    if type == 'wiki':
        print "  * [ticket:%d #%d] (%s) %s" % (int(item[0]), 
            int(item[0]), item[2], item[1])
    else:
        print "  * #%d (%s) %s" % (int(item[0]), item[2], item[1])



#!/usr/bin/env python

# Copyright (C) 2006 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 1, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

import pygtk
pygtk.require('2.0')
import gtk, gobject

import pygst
pygst.require('0.10')
import gst

from gettext import gettext as _
import string, os

import xlmisc, common, media, library 

# some info for every import format (a bit ugly perhaps)
formatdict = {
            "MP3": {
                "Very High":192,
                "High":160,
                "Medium":128,
                "Low":64,
                "ext":"mp3",
                "plugin":"lame"
            },
            "MP3 VBR": {
                "Very High":192,
                "High":160,
                "Medium":128,
                "Low":64,
                "ext":"mp3",
                "plugin":"lame"
            },
            "Ogg Vorbis": {
                "Very High":0.8,
                "High":0.6,
                "Medium":0.4,
                "Low":0.2,
                "ext":"ogg",
                "plugin":"vorbisenc"
            },
            "FLAC": {
                "Very High":8,
                "High": 6,
                "Medium": 4,
                "Low": 2,
                "ext":"flac",
                "plugin":"flacenc"
            }
}


def check_import_formats():
    """
        Return a list of supported encoding formats for importing
    """
    ret = []
    for format, caps in formatdict.iteritems():
        try:
            x = gst.element_factory_find(caps["plugin"])
        except gst.PluginNotFoundError:
            continue
        ret.append(format)

    return ret

class CDImporter(object):
    """
        Imports CDs
    """
    def __init__(self, exaile):
        """
            Initialize 
        """
        self.exaile = exaile
        self.xml = exaile.xml
        self.settings = exaile.settings

        self.importing = False
        self.gst_running = False
        self.overwrite_answer = None

        self.pipeline = None

    def get_pipeline(self, tracknum, dest):
        """
            Construct a gstreamer pipeline either from a list of presets
            or from a string given by the user.
        """
        if self.settings.get_boolean('import/use_custom', False):
            mapping = dict(dest=dest)
            import_custom = self.settings.get_string('import/custom', '')
            template = string.Template(import_custom)
            try:
                cmd = template.substitute(mapping)
            except KeyError, e:
                xlmisc.log('Could not substitute $' + e)
                return
            
            return gst.parse_launch(cmd)
        else:
            format = self.settings.get_str('import/format')
            quality = self.settings.get_str('import/quality')
            vbr = self.settings.get_boolean('import/vbr', False)

            if format == 'MP3':
                encoder = 'lame bitrate=%d ! id3mux v1-tag=1' % (formatdict[format][quality],)
            elif format == 'MP3 VBR':
                encoder = 'lame vbr=4 vbr-mean-bitrate=%d ! id3mux v1-tag=1 v2-tag=1' % \
                    (formatdict[format][quality],)
            elif format == 'Ogg Vorbis':
                encoder = 'vorbisenc quality=%0.1f ! oggmux' % (formatdict[format][quality],)
            elif format == 'FLAC':
                encodern = 'flacenc quality=%d' % (formatdict[format][quality],)
            else:
                xlmisc.log("Unknown import format")
                return None

            source = "cdparanoiasrc track=%d" % (tracknum,)
            sink = "filesink location=\"%s\"" % (dest,)
            cmd = '%s ! audioconvert ! %s ! %s' % (source, encoder, sink)
            try:
                pipe = gst.parse_launch(cmd)
            except Exception, e:
                xlmisc.log("Could not construct gstreamer pipe:" + str(e))

            return pipe

    @common.threaded
    def do_import(self, songs):
        """
            Start importing the given songs
        """
        if not self.exaile.settings.get_str('import/location', '') \
            or not self.exaile.settings.get_str('import/naming', ''):
            gobject.idle_add(common.error, self.exaile.window, 
                            _('Please set the import settings'
                            ' in the Preferences dialog before importing '
                            'anything.'))
            return

        import_xml = gtk.glade.XML('exaile.glade', 'ImportDialog', 'exaile')

        dialog = import_xml.get_widget('ImportDialog')
        dialog.set_transient_for(self.exaile.window)
        dialog.show_all()

        bar = import_xml.get_widget('import_progressbar')
        label = import_xml.get_widget('import_status')
        stop_button = import_xml.get_widget('import_stop_button')
        stop_button.connect('clicked', lambda *e: self.stop_import())

        self.importing = True
        self.overwrite_answer = None

        for i, song in enumerate(songs):
            if not self.importing: break 
            # TRANSLATORS: CD import status
            bar.set_text(_("Track %(current)d/%(total)d") % {
                'current': i + 1,
                'total': len(songs)
            })
            bar.set_fraction(float(i)/len(songs))
            label.set_markup('<big>' + song.title + '</big>')
            self.import_track(song)
            while self.gst_running:
                continue

        dialog.destroy()

        return

    @common.synchronized
    def import_track(self, song):
        """
            Imports one track using a gstreamer pipeline
        """
        song.loc = self.make_dest(song)
        if not song.loc:
            xlmisc.log('Could not construct destination filename for ' + str(song))
        try:
            if not os.path.exists(os.path.dirname(song.io_loc)):
                os.makedirs(os.path.dirname(song.io_loc))
            if os.path.exists(song.io_loc):
                if self.overwrite_answer != "yes_all" \
                    and self.overwrite_answer != "no_all":
                    self.overwrite_answer = None
                    gobject.idle_add(ask_overwrite, self, self.exaile.window, song)
                    while not self.overwrite_answer:
                        pass

                if self.overwrite_answer == "yes" \
                    or self.overwrite_answer == "yes_all":
                    pass
                elif self.overwrite_answer == "no" \
                    or self.overwrite_answer == "no_all":
                    return

        except os.error, e:
                xlmisc.log('Could not create directory "' + 
                    os.path.dirname(song.io_loc) + '": ' + e.args[1])
        
        self.pipeline = self.get_pipeline(song.track, song.io_loc)
        if not self.pipeline:
            xlmisc.log('Could not construct gstreamer pipeline, check your settings!')

        bus = self.pipeline.get_bus()
        bus.add_signal_watch()
        bus.connect('message::error', self.gst_error_cb)
        bus.connect('message::eos', self.gst_eos_cb, song)
        self.gst_running = True
        self.pipeline.set_state(gst.STATE_PLAYING)
        return

    def tag_track(self, track):
        """
            Use the track's metadata to tag the file
        """
        media.write_tag(track)
        return

    def add_track(self, track):
        """
            Add a track to the library
        """
        dir = [os.path.dirname(track.io_loc)]
        self.exaile.library_manager.update_library_add(dir)
        return

    def make_dest(self, song):
        """
            Constructs the destination filename using the 
            metadata and preferences
        """
        location = self.settings.get_str('import/location', '')
        naming = self.settings.get_str('import/naming', '')
        pref = self.settings.get_str('import/format', 'MP3')
        if not pref or not location: 
            return ''

        extension = formatdict[pref]['ext']

        # TODO: this mapping should be more complete (e.g. cd number?)
        mapping = dict(artist=song.artist, album=song.album, title=song.title,\
                    ext=extension, num=(song.track or ''), date=song.date)
        template = string.Template(naming)
        try:
            ret = template.substitute(mapping)
        except KeyError, e:
            xlmisc.log('Could not substitute $' + e)
        ret = os.path.join(location, ret)
        return ret

    def gst_error_cb(self, bus, msg, *args):
        """
            If Gstreamer encounters an error, handle it here
            (Right now all we do is print to the log and stop)
        """
        error = None
        debug = None
        (error, debug) = msg.parse_error()
        xlmisc.log('gstreamer error: ' + error.message)

        self.stop_import()
        return

    def gst_eos_cb(self, bus, msg, track):
        """
        Stop the pipeline when EOS is reached, tag the track and add to library
        """
        self.pipeline.set_state(gst.STATE_NULL)
        self.gst_running = False
        self.tag_track(track)
        self.add_track(track)
        return

    def stop_import(self):
        """
            Stop importing when stop button is clicked
        """
        self.pipeline.set_state(gst.STATE_NULL)
        self.gst_running = False
        self.importing = False



def ask_overwrite(importer, parent, path):
    """
        Throws up a dialog asking whether the user wants to overwrite an 
        existing file, with yes/no/yes to all/no to all alternatives
    """
    dialog = gtk.Dialog(_('Overwrite file'), parent, 
        gtk.DIALOG_MODAL | gtk.DIALOG_DESTROY_WITH_PARENT,
        (_('Yes to All'), gtk.RESPONSE_OK,
        gtk.STOCK_YES, gtk.RESPONSE_YES,
        gtk.STOCK_NO, gtk.RESPONSE_NO,
        _('No to All'), gtk.RESPONSE_CLOSE))

    image = gtk.image_new_from_stock(gtk.STOCK_DIALOG_QUESTION, 
        gtk.ICON_SIZE_DIALOG)
    text = gtk.Label(_('The file %s already exists, do you want to '
            'overwrite it?') % (path,))
    text.set_line_wrap(True)

    hb = gtk.HBox(spacing=8)
    hb.pack_start(image)
    hb.pack_start(text)

    dialog.vbox.pack_start(hb)
    dialog.vbox.show_all()

    response = dialog.run()

    
    if response == gtk.RESPONSE_YES:
        importer.overwrite_answer = "yes" 
    elif response == gtk.RESPONSE_NO:
        importer.overwrite_answer = "no" 
    elif response == gtk.RESPONSE_OK:
        importer.overwrite_answer = "yes_all" 
    elif response == gtk.RESPONSE_CLOSE or \
            response == gtk.RESPONSE_DELETE_EVENT:
        importer.overwrite_answer = "no_all" 

    dialog.destroy()

    return False



# Copyright (C) 2006 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 1, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

import gtk, urllib, gtk.glade, pango, re, md5, os
from gettext import gettext as _
#from urllib import urlencode
from xl import library, xlmisc, common
import xl.path

def update_info(nb, track):
    """
        Changes the information in the info tab if it's already open
    """
    for i in range(0, nb.get_n_pages()):
        page = nb.get_nth_page(i)
        if isinstance(page, TrackInformation):
            page.setup_tabs(track)

def show_information(exaile, track):
    """
        Shows track information
    """
    nb = exaile.playlists_nb
    for i in range(0, nb.get_n_pages()):
        page = nb.get_nth_page(i)
        if isinstance(page, TrackInformation):
            page.setup_tabs(track)
            page.parent.set_current_page(i)
            return page

    return TrackInformation(exaile, track)

class TablatureTab(gtk.VBox):
    """
        Downloads and displays guitar tablature
    """
    REGEX = re.compile("<PRE>(.*)</PRE>", re.DOTALL)
    def __init__(self, panel, track):
        """
            Initializes the panel
        """
        gtk.VBox.__init__(self)
        self.set_border_width(5)
        self.set_spacing(3)
        self.exaile = panel.exaile
        self.panel = panel
        self.stopped = False

        text = gtk.TextView()
        text.set_editable(False)

        text.modify_font(pango.FontDescription("Monospace 11"))
        self.text = text.get_buffer()

        scroll = gtk.ScrolledWindow()
        scroll.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        scroll.add(text)

        self.text.set_text("Downloading tablature...")
        self.fetched = False
        self.pack_start(scroll, True, True)
        
        self.track = track
        self.artist = track.artist.replace(" ", "_").lower()
        self.letter = self.artist[:1]
        self.title = track.title.replace(" ", "_").lower()
        self.url = "http://www.fretplay.com/tabs/%s/%s/%s-tab.shtml" % \
            (self.letter, self.artist, self.title)
        self.cache_file = xl.path.get_cache('tablature_%s.tablature' %
            md5.new(self.url).hexdigest())
        if os.path.isfile(self.cache_file):
            h = open(self.cache_file, 'r')
            self.fetched = True
            self.text.set_text(h.read())
            h.close()
        self.update()
        self.show_all()

    def update(self):
        """
            Called when this tab is selected
        """
        if self.fetched: return
        else:
            self.fetched = False
            xlmisc.log(self.url)
            xlmisc.URLFetcher("www.fretplay.com", "/tabs/%s/%s/%s-tab.shtml" % \
                (self.letter, self.artist, self.title), self.updated).start()

    def updated(self, url, text):
        """
            Called when the text has been fetched.
        """
        if self.track != self.panel.track or \
            self.stopped: return
        m = self.REGEX.search(text)
        if m:
            text = m.group(1)
            text = re.sub("^\s*", "", text)
        else:
            self.text.set_text("Could not fetch tablature.")
        if "Error 302" in text:
            text = "Tablature not found."
        xlmisc.log(self.cache_file)
        if self.track != self.panel.track or \
            self.stopped: return
        self.text.set_text(text)
        h = open(self.cache_file, 'w')
        h.write(text)
        h.close()            

    def close_page(self):
        """
            Called when this tab is closed
        """
        self.stopped = True

class LyricsTab(gtk.VBox):
    """
        Downloads and displays track lyrics
    """
    def __init__(self, exaile, panel, track):
        """
            Initializes the panel
        """
        gtk.VBox.__init__(self)
        self.set_spacing(3)
        self.exaile = exaile
        self.panel = panel
        self.track = track
        self.text = xlmisc.BrowserWindow(exaile, None)#, True) 
        self.pack_start(self.text, True, True)

        self.show_all()
        self.lyrics = self.text
        self.lyrics.set_text("Downloading lyrics, please wait...") 
        self.update()

    def update(self):
        """
            Called when this tab is selected
        """
        params = {'artist': self.track.artist.encode('latin1'), 'songname': self.track.title.encode('latin1')}

#        search = "http://lyrc.com.ar/en/tema1en.php?%s" % urlencode(params)
        search = "http://lyricwiki.org/api.php?artist=%s&song=%s&fmt=html" % (
            urllib.quote_plus(params['artist']), 
            urllib.quote_plus(params['songname']))
        print search
        self.lyrics.load_url(search, self.text.action_count, False)

    def close_page(self):
        """
            Called when this tab is closed
        """
        self.text.stopped = True

class WikipediaTab(gtk.HBox):
    """
        Shows a wikipedia page
    """
    def __init__(self, exaile, url):
        """
            Initializes the panel
        """
        gtk.HBox.__init__(self)
        self.browser = xlmisc.BrowserWindow(exaile, url)
        self.pack_start(self.browser, True, True)
        self.show_all()

    def close_page(self):
        """
            called when this tab is closed
        """
        self.browser.stopped = True

class TrackStatsTab(gtk.ScrolledWindow):
    """
        Track Statistics
    """
    def __init__(self, exaile, track):
        """
            Initializes the tab
        """
        gtk.ScrolledWindow.__init__(self)
        self.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        self.set_border_width(5)

        self.table = table = gtk.Table(12, 2, False) # initialize 12 rows
        table.set_row_spacings(3)
        self.add_with_viewport(table)

        self.db = exaile.db
        self.n_rows = 0

        self.setup_information(track)
        table.resize(self.n_rows, 2)

        self.show_all()

    def setup_information(self, track):
        """
            Adds the specific information fields for this track
        """
        path_id = library.get_column_id(self.db, 'paths', 'name', track.loc)
        row = self.db.read_one("tracks", "plays, rating", "path=?", 
            (path_id,))

        playcount = 0
        rating = 0
        if row:
            playcount = row[0]
            rating = row[1]
        self.append_info(_("Title:"), track.title)
        self.append_info(_("Artist:" ), track.artist)
        self.append_info(_("Album:"), track.album)
        if track.track and track.track > -1:
            self.append_info(_("Track Number:"), str(track.track))
        self.append_info(_("Year:"), track.year)
        self.append_info(_("Genre:"), track.genre)
        self.append_info(_("Bitrate:"), str(track.bitrate))
        if track.rating: self.append_info(_("User Rating:"), track.rating)
        self.append_info(_("System Rating:"), str(rating))
        if playcount >= 0: self.append_info(_("Playcount:"), str(playcount))
        self.append_info(_("Location:"), track.loc.encode(xlmisc.get_default_encoding()))
        self.append_info(_("Date Added:"), track.time_added)

    def append_info(self, label, string):
        """
            Appends an information line
        """
        label = gtk.Label(label)
        label.set_alignment(0, 0)
        attr = pango.AttrList()
        attr.change(pango.AttrWeight(pango.WEIGHT_BOLD, 0, 800))
        label.set_attributes(attr)

        self.table.attach(label, 0, 1, self.n_rows, self.n_rows +1,
            gtk.EXPAND | gtk.FILL, gtk.FILL)

        if isinstance(string, unicode) or isinstance(string, str) or \
            isinstance(string, int):
            label = gtk.Label(unicode(string))
            label.set_alignment(0, 0)
            label.set_line_wrap(True)
            label.set_selectable(True)
            self.table.attach(label, 1, 2, self.n_rows, self.n_rows + 1,
                gtk.EXPAND | gtk.FILL, gtk.FILL)
        else:
            self.table.attach(string, 1, 2, self.n_rows, self.n_rows + 1,
                gtk.EXPAND | gtk.FILL, gtk.FILL)

        self.n_rows += 1;

    def close_page(self):
        """
            Called when this tab is closed
        """
        pass

class RadioTrackStatsTab(TrackStatsTab):
    """
        Track statistic for shoutcase streams
    """
    def __init__(self, exaile, track):
        TrackStatsTab.__init__(self, exaile, track)

    def setup_information(self, track):
        """
            Sets up the specific information fields for this track
        """

        self.append_info(_("Title:"), track.title)
        self.append_info(_("Description:"), track.artist)

        location = gtk.Entry()
        location.set_text(track.loc)
        location.set_editable(False)
        self.append_info(_("Location:"), location)
        self.append_info(_("Bitrate:"), track.bitrate)

class TrackInformation(gtk.VBox):
    """
        Shows information regarding a track
    """
    def __init__(self, exaile, track):
        """
            Initializes the track information page
        """
        self.exaile = exaile
        gtk.VBox.__init__(self)
        self.set_border_width(3)
        top_box = gtk.HBox()
        self.autoswitch = gtk.CheckButton(_("Automatically update on track "
            "change"))
        self.autoswitch.connect('toggled', self.autoswitch_toggled)
        self.autoswitch.set_active(exaile.settings.get_boolean('ui/information_autoswitch',
            False))
        top_box.pack_end(self.autoswitch, False, False)

        self.pack_start(top_box, False, False)
        self.nb = None

        self.db = exaile.db

        self.setup_tabs(track)
        self.show_all()
        self.type = 'information'

        self.exaile.playlists_nb.append_page(self, xlmisc.NotebookTab(exaile,
            _("Information"), self))
        self.exaile.playlists_nb.set_tab_reorderable(self, True)    
        self.exaile.playlists_nb.set_current_page(
            self.exaile.playlists_nb.get_n_pages() - 1)

    def autoswitch_toggled(self, *e):
        """
            Called when the user toggles the autoswitch button
        """
        self.exaile.settings.set_boolean('ui/information_autoswitch', 
            self.autoswitch.get_active())

    def append_page(self, *e):
        """
            Appends a page to the notebook
        """
        self.nb.append_page(*e)

    def setup_tabs(self, track):
        """
            Sets up the tabs for the specified track
        """
        self.track = track
        if self.nb:
            self.remove(self.nb)

        self.nb = gtk.Notebook()
        self.pack_start(self.nb, True, True)

        if track.type == 'stream':
            self.append_page(RadioTrackStatsTab(self.exaile, track),
                gtk.Label(_("Stream Statistics")))
        
        self.append_page(TrackStatsTab(self.exaile, track), gtk.Label(_("Statistics")))
        locale = self.exaile.settings.get_str('wikipedia_locale', 'en')
        
        if xlmisc.mozembed:
            artist = "http://%s.wikipedia.org/wiki/%s" % (locale, track.artist)
            artist = artist.replace(" ", "_")

            self.append_page(WikipediaTab(self.exaile, artist),
                gtk.Label(_("Artist")))
            
            album = "http://%s.wikipedia.org/wiki/%s" % (locale, track.album)
            album = album.replace(" ", "_")
            self.append_page(WikipediaTab(self.exaile, album),
                gtk.Label(_("Album")))
            self.append_page(LyricsTab(self.exaile, self, track),
                gtk.Label(_("Lyrics")))

        else:
            if self.exaile.settings.get_boolean('ui/gnome_extras_warning',
                True):
                dialog = common.ShowOnceMessageDialog(_('Warning'),
                    self.exaile.window, 
                    _('You do not have gnome-python-extras installed.\n'
                    'You will need to install this package (which may have a '
                    'different name in your distribution)\nto use the '
                    '"Artist", "Album", and "Lyrics" tabs'), checked=True)

                result = dialog.run()
                self.exaile.settings.set_boolean('ui/gnome_extras_warning',
                    not result)

            xlmisc.log("gnome-extras not available.  Showing basic"
                       " track information only")

#        self.append_page(TablatureTab(self, track),
#            gtk.Label(_("Tablature")))
        self.show_all()
        self.nb.show_all()

    def close_page(self):
        """
            Called when this tab in the notebook is closed
        """
        for i in range(self.nb.get_n_pages()):
            page = self.nb.get_nth_page(i)
            page.close_page()



# Copyright (C) 2006 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 1, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

import md5, os, random, re, threading, time, traceback, gc, sys
from gettext import gettext as _, ngettext
import gobject, gtk
from xl import common, media, db, audioscrobbler, xlmisc, dbusinterface
from xl.db import DBOperationalError
from xl.gui import library as librarygui

try:    
    import DiscID, CDDB
    CDDB_AVAIL = True
except:
    CDDB_AVAIL = False

def the_cutter(field):
    """
        Cuts "THE " off of the beginning of any field for better sorting
    """
    field = field.lower()
    if field.find("the ") == 0:
        return field.replace("the ", "", 1)
    else:
        return field

def lstrip_special(field):
    """
        Strip special chars off the beginning of a field for sorting. If
        stripping the chars leaves nothing the original field is returned with
        only whitespace removed.
    """
    lowered = field.lower()
    stripped = lowered.lstrip(" `~!@#$%^&*()_+-={}|[]\\\";'<>?,./")
    if stripped:
        return stripped
    return lowered.lstrip()

def get_suggested_songs(exaile, db, song, s, count, done_func):
    """
        Finds and returns 10 suggested songs from last.fm
    """
    new = s[:]
    random.shuffle(new)
    new.insert(0, song)
    new = new[:5]
    all = []

    for song in new:
        if not song.artist: continue
        if song.type == 'stream': continue
        xlmisc.log("Fetching suggested tracks for %s" % song.artist)
        try:
            lastfm = audioscrobbler.AudioScrobblerQuery(artist=song.artist)
            artists = [artist.name for artist in lastfm.similar()]
        except audioscrobbler.AudioScrobblerError:
            continue
        except:
            xlmisc.log_exception()
            continue

        random.shuffle(artists)
        all.extend(artists)
        if len(all) >= 100: break

    gobject.idle_add(done_func, all, count)    

class TrackData:
    """
        Represents a list of tracks
    """
    def __init__(self, tracks=None):
        """
            Initializes the list
        """
        self.total_length = 0
        self.paths = {}
        self._inner = []
        if tracks:
            for track in tracks:
                self.append(track)

    def __getitem__(self, index):
        return self._inner[index]

    def __setitem__(self, index, value):
        old = self._inner[index]
        try:
            del self.paths[old.loc]
        except KeyError:
            pass
        self.paths[value.loc] = value
        self._inner[index] = value
 
    def __len__(self):
        return len(self._inner)

    def index(self, item):
        return self._inner.index(item)
 
    def append(self, track):
        """
            Adds a track to the list
        """
        if not track: return
        self.paths[track.loc] = track
        self._inner.append(track)
        self.update_total_length(track.get_duration(), appending=True)

    def remove(self, track):
        """
            Removes a track from the list
        """
        if not track: return
        try:
            del self.paths[track.loc]
        except KeyError:
            return
        else:
            self._inner.remove(track)
        self.update_total_length(track.get_duration(), appending=False)
    
    def update_total_length(self, track_duration, appending):
        if appending:
            self.total_length += track_duration
        else:
            self.total_length -= track_duration
            
    def get_total_length(self):
        """ 
            Returns length of all tracks in the table as preformatted string
        """
        s = self.total_length
        m, s = divmod(s, 60)
        h, m = divmod(m, 60)
        d, h = divmod(h, 24)

        text = []
        if d: text.append(ngettext("%d day", "%d days", d) % d)
        if h: text.append(ngettext("%d hour", "%d hours", h) % h)
        if m: text.append(ngettext("%d minute", "%d minutes", m) % m)
        if s: text.append(ngettext("%d second", "%d seconds", s) % s)

        text = ", ".join(text)

        return text


    def for_path(self, path):
        """
            Returns the track associated with the path, or None
        """
        return self.paths.get(path)

def search(exaile, all, keyword=None, custom=True):
    """
        Searches tracks for a specified pattern
    """
    if not keyword: return all

    new = TrackData()

    for track in all:
        atts = ('title', 'album', 'artist')

        for attribute in atts:
            v = getattr(track, attribute)
            if v.lower().find(keyword.lower()) > -1:
                if not track in new: new.append(track)
    return new

def search_tracks(parent, db, all, keyword=None, playlist=None, w=None):
    """
        Searches the database for a specific pattern and returns the tracks
        represented by this pattern
    """
    items = []
    args = []
    where = ""
    if keyword and keyword.replace("\"","") == "":
        keyword = None
    if keyword != None and w:
        regex = re.compile("\s+WHERE\s", re.DOTALL)
        w = regex.sub(" AND ", w)
        tokens = keyword.lower().replace("\"","").split(' ')

        anditems = []
        check_fields = ('title', 'artists.name', 'albums.name')
        for token in tokens:
            if not token: continue
            oritems = []
            for field in check_fields:
                oritems.append("%s LIKE \"%%%s%%\"" % (field, token))
            anditems.append('(' + ' OR '.join(oritems) + ')')

        anditems = ' AND '.join(anditems)

        where = 'WHERE (%s) AND ' \
            '(paths.id=tracks.path AND artists.id=tracks.artist AND ' \
            'albums.id=tracks.album)' % anditems

    if w == None:
        if keyword != None or playlist != None:

            if playlist != None:
                rows = db.select("""
                    SELECT 
                        paths.name 
                    FROM playlist_items,paths,playlists
                    WHERE 
                        playlists.name=? AND 
                        playlist_items.playlist=playlists.id AND 
                    paths.id=playlist_items.path
                """, (playlist,))

                for row in rows:
                    items.append(row[0])

                if all:
                    tracks = TrackData()
                    for item in items:
                        tracks.append(all.for_path(item))
                    return tracks
                    
        else:
            new = TrackData() 
            for track in all:
                new.append(track)
            return new

    tracks = TrackData() 

    rows = []
    query = """
        SELECT 
            paths.name 
        FROM tracks,paths,artists,albums %s 
        ORDER BY 
            LOWER(artists.name), 
            LOWER(albums.name), 
            disc_id,
            track, 
            title
        """ % where

    if w != None:
        query = w
        if keyword:
            regex = re.compile("FROM\s+(.*?)\s+AND", re.DOTALL)
            query = regex.sub(r"FROM \1 %s AND" % where, query)


    try:
        cur = db.realcursor()
        cur.execute(query)
    except Exception, e:
        xlmisc.log(query)
        common.error(parent, _("Query Error: %s") % str(e))
        raise e

    while True:
        try:
            row = cur.fetchone()
            if not row: break
        except sqlite3.OperationalError:
            continue

        track = all.for_path(row[0])
        if track == None:
            pass
        else:
            if playlist != None:
                if track.loc in items:
                    tracks.append(track)
            else:
                tracks.append(track)

    cur.close() 
    return tracks

def already_added(t, added):
    """
        Checks to see if the title, artist, album and genre 
        has already been added to the list of tracks
    """

    if dbusinterface.options.testing: return False
    if not t.title: t.title = ""
    if not t.album: t.album = ""
    if not t.artist: t.artist = ""
    if not t.genre: t.genre = ""

    h = "%s - %s - %s - %s" % (t.title, t.album, t.artist, t.genre)

    if added.has_key(h): return True
    added[h] = 1
    return False

def load_tracks(db, current=None):
    """
        Loads all tracks currently stored in the database
    """
    global ALBUMS

    items = ('PATHS', 'ARTISTS', 'RADIO', 'PLAYLISTS')
    for item in items:
        globals()[item] = dict()
    ALBUMS = {}
    added = dict()

    tracks = TrackData()
    for row in db.select("""
        SELECT 
            paths.name, 
            title, 
            artists.name, 
            albums.name, 
            disc_id,
            tracks.genre, 
            track, 
            length, 
            bitrate, 
            year, 
            modified, 
            user_rating, 
            rating,
            blacklisted, 
            time_added,
            encoding,
            plays
        FROM tracks, paths, artists, albums 
        WHERE 
            (
                paths.id=tracks.path AND 
                artists.id = tracks.artist AND 
                albums.id = tracks.album
            ) AND 
            blacklisted=0 
        ORDER BY 
            THE_CUTTER(artists.name), 
            LOWER(albums.name), 
            disc_id,
            track, 
            title
        """):

        t = media.Track(*row)
        path, ext = os.path.splitext(row[0].lower().encode('utf-8'))
        t.type = "file" 

        if already_added(t, added): continue

        tracks.append(t)
    cur = db.cursor(new=True)

    for item in items:
        cur.execute("SELECT id, name FROM %s" % item.lower())
        while True:
            try:
                row = cur.fetchone()
                if not row: break
                globals()[item][row[1]] = row[0]
            except: 
                xlmisc.log_file_and_line()
                xlmisc.log_exception()

    cur.execute("SELECT artist, name, id FROM albums")
    while True:
        try:
            row = cur.fetchone()
            if not row: break
            ALBUMS["%d - %s" % (row[0], row[1])] = row[2]
        except:
            xlmisc.log_exception()

    cur.close()
    db._close_thread()
    return tracks

def scan_dir(dir, files=None, skip=(), exts=(), scanned=None):
    """
        Scans a directory recursively
    """
    regex = None
    if skip:
        match_string = r"^.*(" + r"|".join(skip) + r").*$"
        regex = re.compile(match_string)
    if files is None: 
        files = []
    if scanned is None:
        scanned = set()

    try:
        dir = os.path.realpath(dir)
        to_scan = os.listdir(dir)
    except OSError:
        return files

    if dir in scanned: return files
    scanned.add(dir)

    for file in to_scan:
        try:
            file = os.path.join(dir, file)
        except UnicodeDecodeError:
            xlmisc.log("Error decoding filename %s" % file)
            continue
        except:
            xlmisc.log_exception()
            continue

        try:
            if os.path.isdir(file):
                if regex and regex.match(file): continue
                scan_dir(file, files=files, skip=skip, exts=exts, scanned=scanned)
        except:
            xlmisc.log("Error scanning %s" % file)
            traceback.print_exc()
       
        try:
            (stuff, ext) = os.path.splitext(file)
            if ext.lower() in exts and not file in files:
                files.append(file)
        except:
            traceback.print_exc()
            continue

    return files     

def count_files(directories, skip=()):
    """
        Recursively counts the number of supported files in the specified
        directories
    """
    paths = []
    for dir in directories:
        paths.extend(scan_dir(dir, skip=skip, exts=media.SUPPORTED_MEDIA))

    return paths

@common.threaded
def get_cddb_info(songs, disc_info, exaile):
    """
        Fetches cddb info about an audio cd
    """
    (status, info) = CDDB.query(disc_info)
    if status in (210, 211):
        info = info[0]
        status = 200
    if status != 200: return

    (status, info) = CDDB.read(info['category'], info['disc_id'])
    title = info['DTITLE'].split(" / ")
    for i in range(disc_info[1]):
        songs[i].title = info['TTITLE' + `i`].decode('iso-8859-15',
            'replace')
        songs[i].album = title[1].decode('iso-8859-15', 'replace')
        songs[i].artist = title[0].decode('iso-8859-15', 'replace')
        songs[i].year = info['EXTD'].replace("YEAR: ", "")
        songs[i].genre = info['DGENRE']

    #songs = exaile.tracks.songs
    exaile.tracks.songs = songs

    gobject.idle_add(exaile.tracks.set_songs, songs)

def read_audio_disc(exaile):
    """
        Reads track information from an audio CD
    """
    device = exaile.settings.get_str('cd_device', '/dev/cdrom')
    disc = DiscID.open(device)
    try:
        info = DiscID.disc_id(disc)
    except:
        common.error(exaile.window, _("Could not open audio disc"))
        return None
    minus = 0; total = 0

    songs = TrackData()
    for i in range(info[1]):
        length = ( info[i + 3] / 75 ) - minus
        if i + 1 == info[1]:
            length = info[i + 3] - total

        minus = info[i + 3] / 75
        tracknum = i + 1
        song = media.Track("cdda://%d#%s" % (tracknum, device), _("Track %d") % tracknum,
            track=tracknum, length=length)
        song.type = 'cd'
        #if not song.artist: song.artist = _("Unknown Artist")
        #if not song.album: song.album = _("Unknown Album")
        total += length
        songs.append(song)

    get_cddb_info(songs, info, exaile)

    return songs

@common.synchronized
def get_column_id(db, table, col, value, prep=''):
    """
        Gets a column id for inserting the specific col into the specific
        table
    """
    if not value: value = ''
    cols = globals()['%s%s' % (prep, table.upper())]
    if cols.has_key(value):
        return cols[value]

    vals = cols.values()
    vals.sort()
    if not vals:
        index = 1
    else:
        index = vals[len(vals) - 1]
        index += 1
    cols[value] = index

    db.execute("INSERT INTO %s( id, %s ) VALUES( ?, ? )" % (table, col), 
        (index, value,))
    return index

@common.synchronized
def get_album_id(db, artist_id, album, prep=''):
    cols = globals()['%sALBUMS' % prep]
    if cols.has_key("%d - %s" % (artist_id, album)):
        return cols["%d - %s" % (artist_id, album)]

    vals = cols.values()
    vals.sort()
    if not vals:
        index = 1
    else:
        index = vals[len(vals) - 1]
        index += 1

    cols["%d - %s" % (artist_id, album)] = index
    if not album: album = ''
    db.execute("INSERT INTO albums( id, artist, name ) VALUES( ?, ?, ?)",
        (index, artist_id, album))
    return index

def read_track(db, all, path, track_type=media.Track):
    if not os.path.isfile(path): return None
    if all:
        tr = all.for_path(path)
    else: tr = None
   
    # if the track is not already read, try the database
    if not tr and db:
        tr = read_track_from_db(db, path, track_type)

    # if it's not in the database, read it from the filesystem
    if not tr:
        tr = media.read_from_path(path, track_type)

    return tr

def read_track_from_db(db, path, track_type=media.Track):
    """
        Reads a track from the database
    """
    rows = db.select("""
        SELECT 
            paths.name, 
            title, 
            artists.name, 
            albums.name, 
            disc_id,
            tracks.genre,
            track, 
            length, 
            bitrate, 
            year, 
            modified, 
            user_rating, 
            rating,
            blacklisted, 
            time_added,
            encoding,
            plays
        FROM tracks,paths,artists,albums 
        WHERE 
            (
                paths.id=tracks.path AND 
                artists.id=tracks.artist AND 
                albums.id=tracks.album
            ) 
            AND paths.name=? 
        LIMIT 1
        """, (path,))

    if rows:
        tr = track_type(*rows[0])
    else: 
        tr = None

    if tr and tr.type != 'device':
        tr.type = 'file'
    
    return tr

def save_track_to_db(db, tr, new=False, prep=''):
    if new:
        tr.time_added = time.strftime("%Y-%m-%d %H:%M:%Y", 
            time.localtime())

    path_id = get_column_id(db, 'paths', 'name', tr.loc,
        prep=prep)
    artist_id = get_column_id(db, 'artists', 'name', tr.artist, prep=prep)
    album_id = get_album_id(db, artist_id, tr.album,
        prep=prep)

    db.update("tracks",
        {
            "path": path_id,
            "title": tr.title,
            "artist": artist_id,
            "album": album_id,
            "disc_id": tr.disc_id,
            "genre": tr.genre,
            "track": tr.track,
            "rating": tr.system_rating,
            "user_rating": tr._rating,
            "length": tr.duration,
            "bitrate": tr._bitrate,
            "rating": tr.system_rating,
            "blacklisted": tr.blacklisted,
            "year": tr.year,
            "modified": tr.modified,
            "time_added": tr.time_added,
            "encoding": tr.encoding,
            "plays": tr.playcount
        }, "path=?", (path_id,), new)

class PopulateThread(threading.Thread):
    """
        Reads all the tracks in the library and adds them to the database
    """
    type = 'populate'
    running = False
    stopped = False

    def __init__(self, exaile, db, directories, update_func, 
        delete=True, load_tree=False, done_func=None):
        """
            Expects an exaile instance, the location of the database file,
            the directories to search, the function to call as reading
            progresses, and whether or not this is a quick scan.
        """
        threading.Thread.__init__(self)
        self.db = db
        self.setDaemon(True)
        self.exaile = exaile
        self.directories = directories
        self.update_func = update_func
        self.done = False
        self.delete = delete
        self.load_tree = load_tree
        self.done_func = done_func

    def run(self):
        """
            Called when the thread is started
        """
        xlmisc.log("Running is %s" % PopulateThread.running)
        if PopulateThread.running: return
        PopulateThread.running = True
        PopulateThread.stopped = False

        directories = self.directories
        db = self.db
        cur = self.db.cursor()

        update_func = self.update_func
        gobject.idle_add(update_func, 0.001)

        skip = self.exaile.settings.get_list('scan_ignore',
            default=['incomplete'])
        paths = count_files(directories, skip)
        total = len(paths)
        xlmisc.log("File count: %d" % total)

        if self.type == 'populate':
            db.execute("UPDATE tracks SET included=0")
        count = 0
        update_queue = dict()
        self.added = dict()

        # found_tracks will hold /all/ tracks found in this import, regardless
        # of if they have already been previously imported or not.  They will
        # be handed to the "done_func"
        self.found_tracks = []

        for loc in paths:
            if PopulateThread.stopped:
                self.stop()
                return
            try:
                self.do_function(loc)

            except DBOperationalError:
                xlmisc.log_exception()
                continue
            except OSError:
                xlmisc.log_exception()
                continue
            except Exception, ex:
                xlmisc.log_exception()

            count = count + (1 * 1.0)

            if total != 0 and (count % 3) == 0.0:
                percent = float(count / total)
                gobject.idle_add(update_func, percent)

            # periodical commit
            if count % 1500 == 0:
                xlmisc.log("Committing 1500 scanned tracks...")
                db.commit()

        if PopulateThread.stopped:
            self.stop()
            return
        self.stop()
        xlmisc.log("Count is now: %d" % count)
        if self.done: return

        if self.type == 'populate':
            if self.delete:
                db.execute("DELETE FROM tracks WHERE included=0")
        db.commit()

    def do_function(self, loc):
        db = self.db

        in_library = False
        tr = self.exaile.all_songs.for_path(unicode(loc,
            xlmisc.get_default_encoding()))
        if tr:
            in_library = True

        bl = 0
        if not tr:
            tr = read_track_from_db(db, unicode(loc, xlmisc.get_default_encoding()))
            if tr and tr.blacklisted: bl = 1

        modified = os.stat(loc).st_mtime
        time_added=''
        if not tr or tr.modified != modified:
            if not tr: new = True
            else: 
               new = False
               time_added = tr.time_added
            tr = media.read_from_path(loc)
            tr.blacklisted = bl
            if not tr: return
            if time_added:
               tr.time_added=time_added
            tr.modified = modified
            save_track_to_db(db, tr, new)

        if tr:
            path_id = get_column_id(db, 'paths', 'name', unicode(loc, xlmisc.get_default_encoding()))
            db.execute("UPDATE tracks SET included=1 WHERE path=?",
                (path_id,))
        if not tr or tr.blacklisted: return
        
        if not in_library:
            if not already_added(tr, self.added): self.exaile.all_songs.append(tr)
        self.found_tracks.append(tr)

    def stop(self):
        """
            Stops the thread
        """
        self.db._close_thread()
        self.db.commit()

        num = -2
        if not self.load_tree: num = -1
        tracks = self.found_tracks
        if PopulateThread.stopped:
            tracks = None
        gobject.idle_add(self.update_func, num, tracks, 
            self.done_func) 
        gobject.idle_add(gc.collect)
        PopulateThread.stopped = False
        PopulateThread.running = False
        self.exaile = None
        self.directories = None
        self.added = None

def populate(exaile, db, directories, update_func, delete=True,
    load_tree=False, done_func=None):
    """
        Runs the populate thread
    """
    thread = PopulateThread(exaile, db, directories, update_func, 
        delete, load_tree, done_func)
    exaile.thread_pool.append(thread)
    thread.start()

class RemoveTracksThread(PopulateThread):
    """
        Removes tracks in database according to directories given
    """
    type = 'remove'
    def __init__(self, exaile, db, directories, update_func, 
        delete=True, load_tree=False, done_func=None):
        """
            Expects an exaile instance, the location of the database file,
            the directories to search, the function to call as reading
            progresses, and whether or not this is a quick scan.
        """
        PopulateThread.__init__(self, exaile, db, directories, update_func,
            delete, load_tree, done_func)

    def do_function(self, loc):
        db = self.db
        tr = self.exaile.all_songs.for_path(loc)
        
        if not tr:
            return

        if tr:
            path_id = get_column_id(db, 'paths', 'name', unicode(loc, xlmisc.get_default_encoding()))
            db.execute("DELETE FROM tracks WHERE path=?",
                (path_id,))

        try:
            self.exaile.all_songs.remove(tr)
        except ValueError:
            pass

def remove_tracks(exaile, db, directories, update_func, delete=True,
    load_tree=False, done_func=None):
    """
       Removes tracks in database specified by directory path(s)
    """
    thread = RemoveTracksThread(exaile, db, directories, update_func,
        delete, load_tree, done_func)
    exaile.thread_pool.append(thread)
    thread.start()

class AddTracksThread(PopulateThread):
    """
        Adds tracks to database according to directories given
    """
    type = 'add'
    def __init__(self, exaile, db, directories, update_func, 
        delete=True, load_tree=False, done_func=None):
        """
            Expects an exaile instance, the location of the database file,
            the directories to search, the function to call as reading
            progresses, and whether or not this is a quick scan.
        """
        PopulateThread.__init__(self, exaile, db, directories, update_func,
            delete, load_tree, done_func)

    def do_function(self, loc):
        db = self.db

        if self.exaile.all_songs.for_path(loc): return
        tr = media.read_from_path(loc)
        if not tr: return
        new = True
        tr.modified = os.stat(loc).st_mtime
        save_track_to_db(db, tr, new)
        path_id = get_column_id(db, 'paths', 'name', unicode(loc, xlmisc.get_default_encoding()))
        db.execute("UPDATE tracks SET included=1 WHERE path=?", (path_id,))
        
        if not already_added(tr, self.added): self.exaile.all_songs.append(tr)
        self.found_tracks.append(tr)

def add_tracks(exaile, db, directories, update_func, delete=True,
    load_tree=False, done_func=None):
    """
       Add tracks in database specified by directory path(s)
    """
    thread = AddTracksThread(exaile, db, directories, update_func,
        delete, load_tree, done_func)
    exaile.thread_pool.append(thread)
    thread.start()

class LibraryManager(object):
    def __init__(self, exaile):
        self.exaile = exaile
        self.settings = exaile.settings
        self.db = exaile.db

    def import_directory(self, load_tree=False):
        """
            Imports a single directory into the database
        """
        dialog = gtk.FileChooserDialog(_("Add a directory"),
            self.exaile.window, gtk.FILE_CHOOSER_ACTION_SELECT_FOLDER,
            (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
            gtk.STOCK_ADD, gtk.RESPONSE_OK))
        dialog.set_current_folder(self.exaile.get_last_dir())

        checkbtn = gtk.CheckButton(_("Add tracks to current playlist after importing"))
        dialog.set_extra_widget(checkbtn)

        items = []
        tmp = self.settings.get_list("search_paths")
        for i in tmp:
            if i != "": items.append(i)

        response = dialog.run()
        if response == gtk.RESPONSE_OK:
            path = dialog.get_filename()

            check = True
            for p in items:
                if p == path: check = False
                if p.find(path) > -1: check = False
                if path.find(p) > -1: check = False

            if check:
                items.append(path)

            self.settings['search_paths'] = items

            done_func = None
            if checkbtn.get_active():
                done_func = self.after_import
            self.update_library((path,), done_func=done_func, load_tree=load_tree, 
                delete=False)
        dialog.destroy()

    def after_import(self, songs):
        """
            Adds songs that have just been imported to the current playlist
            after importing a directory
        """
        songs = self.exaile.tracks.reorder_songs(songs)
        self.exaile.playlist_manager.append_songs(songs, play=False)

    def show_library_manager(self):
        """
            Displays the library manager
        """
        dialog = librarygui.LibraryDialog(self.exaile)
        response = dialog.run()
        dialog.dialog.hide()
        if response == gtk.RESPONSE_APPLY:
            self.on_library_remove_tracks()
            self.on_library_add_tracks()
        dialog.destroy()

    @common.threaded
    def load_songs(self, updating=False, first_run=False): 
        """
            Loads the entire library from the database
        """
        gobject.idle_add(self.exaile.status.set_first, 
            _("Loading library from database..."))

        if not updating:
            xlmisc.log("loading tracks...")
            self.exaile.all_songs = load_tracks(self.db, 
                self.exaile.all_songs)
            self.db._close_thread()
            xlmisc.log("done loading tracks...")
        gobject.idle_add(self.exaile.status.set_first, None)

        self.exaile.collection_panel.songs = self.exaile.all_songs
        self.exaile.collection_panel.track_cache = dict()

        if not updating:
            xlmisc.log("loading songs")
            gobject.idle_add(self.exaile.playlists_panel.load_playlists)
            gobject.idle_add(self.exaile.collection_panel.load_tree, True)
            
        if first_run: 
            gobject.idle_add(self.exaile.initialize)

    def on_library_rescan(self, widget=None, event=None, data=None,
        load_tree=True): 
        """
            Rescans the library for newly added tracks
        """
        xlmisc.log('Library rescan called')
        items = []
        tmp = self.settings.get_list("search_paths", [])
        for i in tmp:
            if i != "": items.append(i)

        if len(items): self.update_library(items, load_tree=load_tree)
    
    def update_library(self, items, done_func=None,
        load_tree=True, delete=True): 
        """
            Updates the library
        """
        self.exaile.status.set_first(_("Scanning collection..."))

        populate(self.exaile, self.db,
            items, self.on_library_update, delete,
            load_tree=load_tree, done_func=done_func)

    def on_library_update(self, percent, songs=None, done_func=None): 
        """
            Scans the library
        """
        self.exaile.collection_panel.update_progress(percent)

        if percent < 0:
            self.db.db.commit()
            self.db._cursor.close()
            self.db._cursor = self.db.realcursor()

            # if percent is not -2, it wasn't an auto scan, so we won't reload the
            # collection tree because that could potentially take a lot of
            # time if the person has a lot of tracks in their library
            if percent == -2:
                self.exaile.collection_panel.songs = self.exaile.all_songs
                self.exaile.collection_panel.track_cache = dict()
                gobject.idle_add(self.exaile.collection_panel.load_tree, True)

        if done_func:
            done_func(songs)

    def on_library_add_tracks(self, widget=None, event=None, data=None,
        load_tree=True): 
        """
            Add tracks from paths added to library manager
        """
        items = []
        tmp = self.settings.get_list("add_paths", [])
        for i in tmp:
            if i != "": items.append(i)

        if len(items): self.update_library_add(items, load_tree=load_tree)

    def update_library_add(self, items, done_func=None,
        load_tree=True, delete=True): 
        """
            Updates the library by adding tracks
        """
        self.exaile.status.set_first(_("Adding tracks..."))

        add_tracks(self.exaile, self.db,
            items, self.on_library_update, delete,
            load_tree=load_tree, done_func=done_func)

    def on_library_remove_tracks(self, widget=None, event=None, data=None,
        load_tree=True): 
        """
            Remove tracks from paths deleted from library manager
        """
        items = []
        tmp = self.settings.get_list("remove_paths", [])
        for i in tmp:
            if i != "": items.append(i)

        if len(items): self.update_library_remove(items, load_tree=load_tree)

    def update_library_remove(self, items, done_func=None,
        load_tree=True, delete=True): 
        """
            Updates the library by removing tracks
        """
        self.exaile.status.set_first(_("Removing tracks..."))

        remove_tracks(self.exaile, self.db,
            items, self.on_library_update, delete,
            load_tree=load_tree, done_func=done_func)



# Copyright (C) 2006 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 1, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

import gtk, os, gobject, urllib
from gettext import gettext as _, ngettext
from xl import common, media, library, xlmisc
from xl.gui import editor, playlist
import xl.path

class CollectionPanel(object):
    """
        Represents the entire collection in a tree
    """
    rating_images = []
    rating_width = 64   # some default value
    old_r_w = -1
    name = 'col'
    def __init__(self, exaile):
        """
            Initializes the collection panel. Expects a parent and exaile
            object
        """

        # FIXME: Refactor get_selected_items to be get_selected_tracks
        self.get_selected_tracks = self.get_selected_items

        self.xml = exaile.xml
        self.exaile = exaile
        self.db = exaile.db

        self.model = gtk.TreeStore(gtk.gdk.Pixbuf, object, str)
        self.model_blank = gtk.TreeStore(gtk.gdk.Pixbuf, object, str)
        self.scan_label = None
        self.queue_item = -1
        self.scan_progress = None
        self.showing = False
        self.tree = None
        self.keyword = None
        self.track_cache = dict()
        self.start_count = 0
        self.artist_image = gtk.gdk.pixbuf_new_from_file(xl.path.get_data(
            'images', 'artist.png'))
        self.year_image = gtk.gdk.pixbuf_new_from_file(xl.path.get_data(
            'images', 'year.png'))
        self.album_image = self.exaile.window.render_icon('gtk-cdrom',
            gtk.ICON_SIZE_SMALL_TOOLBAR)
        self.track_image = gtk.gdk.pixbuf_new_from_file(xl.path.get_data(
            'images', 'track.png'))
        self.genre_image = gtk.gdk.pixbuf_new_from_file(xl.path.get_data(
            'images', 'genre.png'))
        self.iplaylist_image = gtk.gdk.pixbuf_new_from_file(xl.path.get_data(
            'images', 'playlist.png'))
        self.connect_id = None
        self.setup_widgets()
        playlist.create_rating_images(self)
        self.cover_cache = {}

    def setup_widgets(self):
        """
            Sets up the widgets for this panel
        """
        self.box = self.xml.get_widget('%s_box' % self.name)
        self.choice = self.xml.get_widget('%s_combo_box' % self.name)

        active = self.exaile.settings.get_int('ui/%s_active_view' % self.name, 0)
        self.choice.set_active(active)
        self.xml.get_widget('%s_refresh_button' % self.name).connect('clicked',
            self.load_tree)
        self.xml.get_widget('%s_combo_box' % self.name).connect('changed',
            self.load_tree)
        self.filter = xlmisc.ClearEntry(self.key_release)
        self.xml.get_widget('%s_filter_box' % 
            self.name).pack_start(self.filter.entry,
            True, True)
        self.filter.connect('activate', self.on_search)
        self.key_id = None
        self.filter.set_sensitive(False)

    def key_release(self, *e):
        """
            Called when someone releases a key.
            Sets up a timer to simulate live-search
        """
        if self.key_id:
            gobject.source_remove(self.key_id)
            self.key_id = None

        self.key_id = gobject.timeout_add(700, self.on_search)

    def update_progress(self, percent):
        """
            Updates scanning progress meter
        """
        if percent <= -1:
            self.scan_label.hide()
            self.progress_box.hide()
            self.progress.hide()
            self.stop_button.hide()
            self.showing = False
            self.exaile.status.set_first(_("Finished scanning collection."),
                2000)
            self.scan_label.destroy()
            self.progress_box.destroy()
            self.stop_button.destroy()
            self.progress.destroy()
            self.scan_label = None

        else:
            if not self.scan_label:
                self.progress = gtk.ProgressBar()
                self.progress_box = gtk.HBox(spacing=2)
                self.stop_button = gtk.Button()
                image = gtk.Image()
                image.set_from_stock('gtk-stop', gtk.ICON_SIZE_SMALL_TOOLBAR)
                self.stop_button.set_image(image)
                self.stop_button.connect('clicked', self.stop_scan)
                self.progress_box.pack_start(self.progress, True, True)
                self.progress_box.pack_start(self.stop_button, False, False)

                self.box.pack_end(self.progress_box, False, False)
                # TRANSLATORS: Scanning filesystem for tracks
                self.scan_label = gtk.Label(_("Scanning..."))
                self.scan_label.set_alignment(0, 0)
                self.box.pack_end(self.scan_label, False, False)
           
            if not self.showing:
                self.showing = True
                self.progress.show()
                self.stop_button.show()
                self.progress_box.show_all()
                self.scan_label.show()
            self.progress.set_fraction(percent)

    def stop_scan(self, widget):
        """
            Stops the library scan
        """
        if library.PopulateThread.running:
            library.PopulateThread.stopped = True

    def on_search(self, widget=None, event=None):
        """
            Searches tracks and reloads the tree.  A timer is started
            in LoadTree if keyword is not None which will expand the items
            of the tree until the matched node is found
        """
        self.keyword = unicode(self.filter.get_text(), 'utf-8')
        self.start_count += 1
        self.load_tree()

    def create_popup(self):
        """
            Creates the popup menu for this tree
        """
        menu = xlmisc.Menu()
        # TRANSLATORS: Append selected track(s) to current playlist
        self.append = menu.append(_("Append to Current"),
            self.append_to_playlist, 'gtk-add')

        pm = xlmisc.Menu()
        self.new_playlist = pm.append(_("New Playlist"),
            self.add_items_to_playlist, 'gtk-new')
        pm.append_separator()

        rows = self.db.select("SELECT name FROM playlists WHERE type=0 ORDER BY"
            " name")

        for row in rows:
            pm.append(row[0], self.add_to_playlist)

        menu.append_menu(_("Add to Playlist"), pm, 'gtk-add')

        pixbuf = xlmisc.get_text_icon(self.exaile.window, u'\u2610', 16, 16)
        icon_set = gtk.IconSet(pixbuf)
        
        factory = gtk.IconFactory()
        factory.add_default()        
        factory.add('exaile-queue-icon', icon_set)

        self.queue_item = menu.append(_("Queue Items"),
            self.append_to_playlist, 'exaile-queue-icon')
        menu.append_separator()
        # TRANSLATORS: Put selected track(s) on the blacklist
        self.blacklist = menu.append(_("Blacklist Selected"),
            self.remove_items, 'gtk-delete')
        # TRANSLATORS: Remove selected track(s) from the disk
        self.remove = menu.append(_("Delete Selected"), 
            self.remove_items, 'gtk-delete')

        n_selected = len(self.get_selected_tracks())
        menu.append_separator()
        em = xlmisc.Menu()

        em.append(_("Edit Information"), lambda e, f:
            editor.TrackEditor(self.exaile, self), 'gtk-edit')
        em.append_separator()

        rm = xlmisc.Menu()

        for i in range(0, 5):
            if i == 0:
                item = rm.append('-', lambda w, e, i=i:
                    editor.update_rating(self, i))
            else:
                item = rm.append_image(self.rating_images[i],
                    lambda w, e, i=i: editor.update_rating(self, i))

        em.append_menu(_("Rating"), rm)
        menu.append_menu(ngettext("Edit Track", "Edit Tracks", n_selected), em,
            'gtk-edit')

        self.menu = menu

    def add_to_playlist(self, widget, event):
        """
            Adds items to the playlist
        """
        playlist = widget.get_child().get_label()
        items = self.get_selected_items()
        self.exaile.playlists_panel.add_items_to_playlist(playlist, items)

    def drag_get_data(self, treeview, context, selection, target_id, etime):
        """
            Called when a drag source wants data for this drag operation
        """
        urls = self.get_urls_for(self.get_selected_items())
        selection.set_uris(urls)

    def get_urls_for(self, items): # may be overridden
        """
            Returns the the items' URLs
        """
        return [urllib.quote(item.loc.encode(xlmisc.get_default_encoding()))
            for item in items]

    def append_recursive(self, iter, add):
        """
            Appends items recursively to the added songs list.  If this
            is a genre, artist, or album, it will search into each one and
            all of the tracks contained
        """
        iter = self.model.iter_children(iter)        
        while True:
            field = self.model.get_value(iter, 2)
            if self.model.iter_has_child(iter):
                self.append_recursive(iter, add)
            elif field == 'title':
                track = self.model.get_value(iter, 1)
                if not track.loc in add:
                    add.append(track.loc)
            
            iter = self.model.iter_next(iter)
            if not iter: break

    def get_selected_items(self):
        """
            Finds all the selected tracks
        """

        selection = self.tree.get_selection()
        (model, paths) = selection.get_selected_rows()
        found = [] 
        for path in paths:
            iter = self.model.get_iter(path)
            field = self.model.get_value(iter, 2)
            if self.model.iter_has_child(iter):
                self.append_recursive(iter, found)
            else:
                track = self.model.get_value(iter, 1)
                if field == 'title':
                    if not track.loc in found:
                        found.append(track.loc)

        add = library.TrackData()
        for row in found:
            add.append(self.all.for_path(row))

        return add

    def remove_items(self, item, event):
        """
            Removes or blacklists tracks
        """

        add = self.get_selected_items() 
        device_delete = []

        if item == self.remove:
            result = common.yes_no_dialog(self.exaile.window,
                _("Are you sure you want to permanently remove the " \
                "selected tracks from disk?"))

            if result != gtk.RESPONSE_YES: return

            for track in add:
                if track.type == 'device':
                    device_delete.append(track)
                    continue
                else:
                    for track in add:
                        os.remove(track.loc)
                path_id = library.get_column_id(self.db, 'paths', 'name',
                    track.loc)
                self.db.execute("DELETE FROM tracks WHERE path=?", (path_id,))
                self.db.execute("DELETE FROM playlist_items WHERE path=?",(path_id,))

        if item == self.blacklist:
            result = common.yes_no_dialog(self.exaile.window,
                _("Are you sure you want to blacklist the selected tracks?"))
            if result != gtk.RESPONSE_YES: 
                return
            for track in add:
                if track.type == 'device':
                    continue
                path_id = library.get_column_id(self.db, 'paths', 'name',
                    track.loc)
                self.db.execute("UPDATE tracks SET blacklisted=1 WHERE path=?", (path_id,))
            
        for track in add:
            try: self.exaile.all_songs.remove(track)
            except: pass
            try: self.exaile.songs.remove(track)
            except: pass
            try: self.exaile.playlist_songs.remove(track)
            except: pass

        if device_delete:   
            self.remove_tracks(device_delete)

        if self.exaile.tracks: 
            self.exaile.tracks.set_songs(self.exaile.songs)
        self.track_cache = dict()
        self.load_tree()

    def add_items_to_playlist(self, item=None, *e):

        add = self.get_selected_items()
        self.exaile.playlists_panel.on_add_playlist(item, None, add)

    def append_to_playlist(self, item=None, event=None):
        add = self.get_selected_items()
        queue = (item == self.queue_item)

        self.exaile.playlist_manager.append_songs(add, queue, True)

    def button_release(self, widget, event):
        """
            Called when a button is released
        """
        if event.button != 1 or self.__dragging: return True
        if event.state & (gtk.gdk.SHIFT_MASK|gtk.gdk.CONTROL_MASK):
            return True
        selection = self.tree.get_selection()
        x, y = event.get_coords()
        x = int(x); y = int(y)

        path = self.tree.get_path_at_pos(x, y)
        if not path: return False
        selection.unselect_all()
        selection.select_path(path[0])

    def button_press(self, widget, event):
        """
            Called when someone clicks on the tree
        """
        selection = self.tree.get_selection()

        (x, y) = event.get_coords()
        x = int(x); y = int(y)
        path = self.tree.get_path_at_pos(x, y)
        if event.type == gtk.gdk._2BUTTON_PRESS:
            (model, paths) = selection.get_selected_rows()

            # check to see if it's a double click on an album
            if len(paths) == 1:
                iter = self.model.get_iter(path[0])
                object = self.model.get_value(iter, 1)
                field = self.model.get_value(iter, 2)

                # if this is a device panel, check to see if the current
                # driver wants to handle this object, and if so, return
                if self.name == 'device':
                    if self.driver and hasattr(self.driver, 'check_open_item'):
                        if self.driver.check_open_item(object):
                            return False
                if field == 'album':
                    self.append_to_playlist()
                    return False

            for path in paths:
                iter = self.model.get_iter(path)
                object = self.model.get_value(iter, 1)
                if self.model.iter_has_child(iter):
                    self.tree.expand_row(path, False)
                else:
                    self.append_to_playlist() 
            return False

        iter = self.model.get_iter(path[0])
        object = self.model.get_value(iter, 1)

        self.create_popup()

        if self.name == 'device':
            self.show_device_panel_menu(widget, event, object)
            return True

        self.menu.popup(None, None, None, event.button, event.time)
        if selection.count_selected_rows() <= 1: return False
        return True

    def track_data_func(self, column, cell, model, iter, user_data=None):
        """
            Called when the tree needs a value for column 1
        """
        object = model.get_value(iter, 1)
        if object is None: return
        field = model.get_value(iter, 2)

        if field == 'nofield':
            cell.set_property('text', object)
            return
        if hasattr(object, field):
            info = getattr(object, field)
            if not info: info = _('Unknown')
            cell.set_property('text', info)

    def drag_data_received(self, tv, context, x, y, selection, info, etime):
        pass

    def get_initial_root(self, model):
        """
            gets the initial root node
        """
        return None

    def load_tree(self, event=None):
        """
            Builds the tree.  If self.keyword is not None, it will start a timer
            that will expand each node until a node that matches the keyword
            is found
        """
        self.current_start_count = self.start_count
        if not self.tree:
            self.tree = xlmisc.DragTreeView(self)
            self.tree.set_headers_visible(False)
            container = self.xml.get_widget('%s_box' % self.name)
            scroll = gtk.ScrolledWindow()
            scroll.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
            scroll.add(self.tree)
            scroll.set_shadow_type(gtk.SHADOW_IN)
            container.pack_start(scroll, True, True)
            container.show_all()

            selection = self.tree.get_selection()
            selection.set_mode(gtk.SELECTION_MULTIPLE)
            pb = gtk.CellRendererPixbuf()
            cell = gtk.CellRendererText()
            col = gtk.TreeViewColumn('Text')
            col.pack_start(pb, False)
            col.pack_start(cell, True)
            col.set_attributes(pb, pixbuf=0)
            self.tree.append_column(col)
            col.set_cell_data_func(cell, self.track_data_func)

            self.tree.set_row_separator_func(
                lambda m, i: m.get_value(i, 1) is None)

        # clear out the tracks if this is a first time load or the refresh
        # button is pressed
        if event: 
            xlmisc.log("Clearing tracks cache")
            self.track_cache = dict()

        self.model.clear()
        self.tree.set_model(self.model_blank)
        self.root = self.get_initial_root(self.model)
    
        self.image_map = {
            "album": self.album_image,
            "artist": self.artist_image,
            "genre": self.genre_image,
            "title": self.track_image,
            "year": self.year_image,
        }

        if self.keyword == "": self.keyword = None

        orders = (
            ('artist', 'album', 'track', 'title'),
            ('album', 'track', 'title'),
            ('genre', 'artist', 'album', 'track', 'title'),
            ('genre', 'album', 'artist', 'track', 'title'),
            ('year', 'artist', 'album', 'track', 'title'),
            ('year', 'album', 'artist', 'track', 'title'),
            ('artist', 'year', 'album', 'track', 'title')
        )
        self.order = orders[self.choice.get_active()]

        o_map = {
            'album': 'LSTRIP_SPEC(albums.name), disc_id', 
            'artist': 'LSTRIP_SPEC(THE_CUTTER(artists.name))',
            'genre': 'LSTRIP_SPEC(genre)',
            'title': 'LSTRIP_SPEC(title)',
            'track': 'track', 
            'year' : 'LSTRIP_SPEC(year)',
        }
        order_by = ', '.join((o_map[o] for o in self.order))

        self.where = """
            SELECT 
                paths.name 
            FROM tracks, albums, paths, artists
            WHERE 
                blacklisted=0 AND
                (
                    paths.id=tracks.path AND
                    albums.id=tracks.album AND
                    artists.id=tracks.artist 
                )
            ORDER BY 
                %s
            """ % order_by


        # save the active view setting
        self.exaile.settings['ui/%s_active_view' % self.name] = self.choice.get_active()

        # if this is a collection panel, don't alter self.all, as this messes
        # up inheriting from it
        if self.name == 'col':
            all = self.exaile.all_songs
            self.all = all
        else:
            # FIXME: This is evil OOP; it should use method overriding instead,
            # e.g. see drag_get_data and get_urls_for.
            if isinstance(self, device.DevicePanel):
                self.all = None

        songs = None
        key = "%s %s" % (self.where, self.keyword)
        try:
            songs = self.track_cache[key]
        except:
            pass

        if not songs or self.name== 'device':
            songs = self.search_tracks(self.keyword, self.all)
            self.track_cache[key] = songs

        if self.current_start_count != self.start_count: return

        self.append_info(self.root, songs)

        if self.connect_id: gobject.source_remove(self.connect_id)
        self.connect_id = None
        self.filter.set_sensitive(True)

    def search_tracks(self, keyword, all):
        """
            Searches for songs
        """
        return library.search_tracks(self.exaile.window, self.db, all,
            self.keyword, None, self.where)

    def __check_track_function(self, model, path, iter, track_needed):
        """
           Checks if track is equal to given while traversing the tree,
           called by show_in_collection
        """
        node = self.model.get_value(iter, 1)
        if isinstance(node, media.Track) and node.loc == track_needed.loc:
            self.__track_path = path
            return True

    def show_in_collection(self, track):
        """
           Shows given track in collection panel:
           1. Expand tree to this node
           2. Scroll tree to centerize node vertically if possible
           3. Set cursor to it
        """
        self.__track_path = None
        self.model.foreach(self.__check_track_function, track)
        path = self.__track_path
        if path is None:
            common.info(self.exaile.window,
                _("Track is not present in collection"))
        else:
            self.tree.set_model(self.model)
            self.tree.expand_to_path(path)
            gobject.idle_add(self.tree.scroll_to_cell, path, None, True, 0.5)
            gobject.idle_add(self.tree.set_cursor, path)

    @common.threaded
    def append_info(self, node, songs=None, unknown=False):
        """
            Appends all related info and organizes the tree based on self.order
            Only the very last item of self.order will be a child node
        """

        show_covers = self.exaile.settings.get_boolean('ui/covers_in_collection', False)
        order_nodes = common.idict()
        order = []
        last_songs = []
        for field in self.order:
            if field == "track": continue
            order.append(field)

        for field in order:
            order_nodes[field] = common.idict()

        expanded_paths = []
        last_char = None
        use_alphabet = self.exaile.settings.get_boolean("ui/use_alphabet", True)
        for track in songs:
            if self.current_start_count != self.start_count: return

            parent = node
            last_parent = None
            string = ""
            first = True
            for field in order:
                node_for = order_nodes[field]
                if field == "track": continue
                info = getattr(track, field)

                # print separators
                if first and info and use_alphabet:
                    temp = library.the_cutter(library.lstrip_special(info).upper()).upper()

                    if not temp:
                        first_char = ' '
                    else:
                        first_char = temp[0]
                    if not last_char: # First row, don't add separator.
                        last_char = first_char
                    if first_char != last_char:
                        if not first_char.isalpha():
                            first_char = '0-9'
                        if first_char != last_char:
                            last_char = first_char
                            self.model.append(parent, [None, None, None]) # separator

                if not info: 
                    if not unknown and first:
                        last_songs.append(track)
                        break
                    info = _("Unknown")
                first = False

                if field == "title":
                    n = self.model.append(parent, [self.track_image,
                        track, field])
                else:
                    string = "%s - %s" % (string, info)
                    if not string in node_for:
                        if field == "album" and show_covers:                                            
                            album_cover = self.exaile.cover_manager.fetch_cover(track, True)
                            if not album_cover in self.cover_cache:
                                self.cover_cache[album_cover] = \
                                gtk.gdk.pixbuf_new_from_file_at_size(album_cover, 20, 20)
                            parent = self.model.append(parent, 
                                [self.cover_cache[album_cover], track, field])
                        else:                            
                            parent = self.model.append(parent, 
                                [self.image_map[field], track, field])

                        if info == "track": info = track
                        node_for[string] = parent
                    else: parent = node_for[string]

                if self.keyword and last_parent:
                    if self.keyword.lower() in common.to_unicode(info).lower():
                        expanded_paths.append(self.model.get_path(
                            last_parent))
                last_parent = parent

        # make sure "Unknown" items end up at the end of the list
        if not unknown and last_songs:
            if use_alphabet:
                self.model.append(node, [None, None, None]) # separator
            self.append_info(self.root, last_songs, True)

        gobject.idle_add(self.tree.set_model, self.model)
        for path in expanded_paths:
            gobject.idle_add(self.tree.expand_to_path, path)

        if self.root:
            path = self.model.get_path(self.root)
            gobject.idle_add(self.tree.expand_row, path, False)
            if self.name == 'device':
                gobject.idle_add(self.done_loading_tree)

# HACK: Work around failing recursive import.
# See fixme in load_tree for why this is needed and how it should be fixed.
import device
